#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import numpy as np
from typing import Optional, List

from nomad.units import ureg
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.numerical_settings import (
    KMesh,
    KLinePath,
    KSpaceFunctionalities,
)

from . import logger
from .conftest import generate_k_line_path, generate_k_space_simulation


class TestKSpace:
    """
    Test the `KSpace` class defined in `numerical_settings.py`.
    """

    @pytest.mark.parametrize(
        'system_type, is_representative, reciprocal_lattice_vectors, result',
        [
            ('bulk', False, None, None),
            ('atom', True, None, None),
            ('bulk', True, None, [[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
            (
                'bulk',
                True,
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            ),
        ],
    )
    def test_normalize(
        self,
        system_type: Optional[str],
        is_representative: bool,
        reciprocal_lattice_vectors: Optional[List[List[float]]],
        result: List[List[float]],
    ):
        """
        Test the `normalize` method. This also test the `resolve_reciprocal_lattice_vectors` method.
        """
        simulation = generate_k_space_simulation(
            system_type=system_type,
            is_representative=is_representative,
            reciprocal_lattice_vectors=reciprocal_lattice_vectors,
        )
        k_space = simulation.model_method[0].numerical_settings[0]
        assert k_space.name == 'KSpace'
        k_space.normalize(EntryArchive(), logger)
        if k_space.reciprocal_lattice_vectors is not None:
            value = k_space.reciprocal_lattice_vectors.to('1/angstrom').magnitude / (
                2 * np.pi
            )
            assert np.allclose(value, result)
        else:
            assert k_space.reciprocal_lattice_vectors == result


class TestKSpaceFunctionalities:
    """
    Test the `KSpaceFunctionalities` class defined in `numerical_settings.py`.
    """

    @pytest.mark.parametrize(
        'reciprocal_lattice_vectors, check_grid, grid, result',
        [
            (None, None, None, False),
            ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], False, None, True),
            ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], True, None, False),
            ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], True, [6, 6, 6, 4], False),
            ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], True, [6, 6, 6], True),
        ],
    )
    def test_validate_reciprocal_lattice_vectors(
        self,
        reciprocal_lattice_vectors: Optional[List[List[float]]],
        check_grid: bool,
        grid: Optional[List[int]],
        result: bool,
    ):
        """
        Test the `validate_reciprocal_lattice_vectors` method.
        """
        check = KSpaceFunctionalities().validate_reciprocal_lattice_vectors(
            reciprocal_lattice_vectors=reciprocal_lattice_vectors,
            logger=logger,
            check_grid=check_grid,
            grid=grid,
        )
        assert check == result

    def test_resolve_high_symmetry_points(self):
        """
        Test the `resolve_high_symmetry_points` method. Only testing the valid situation in which the `ModelSystem` normalization worked.
        """
        # `ModelSystem.normalize()` need to extract `bulk` as a type.
        simulation = generate_k_space_simulation(
            pbc=[True, True, True],
        )
        model_systems = simulation.model_system
        # normalize to extract symmetry
        simulation.model_system[0].normalize(EntryArchive(), logger)

        # Testing the functionality method
        high_symmetry_points = KSpaceFunctionalities().resolve_high_symmetry_points(
            model_systems=model_systems, logger=logger
        )
        assert len(high_symmetry_points) == 4
        assert high_symmetry_points == {
            'Gamma': [0, 0, 0],
            'M': [0.5, 0.5, 0],
            'R': [0.5, 0.5, 0.5],
            'X': [0, 0.5, 0],
        }


class TestKMesh:
    """
    Test the `KMesh` class defined in `numerical_settings.py`.
    """

    @pytest.mark.parametrize(
        'center, grid, result_points, result_offset',
        [
            # No `center` and `grid`
            (None, None, None, None),
            # No `grid`
            ('Gamma-centered', None, None, None),
            ('Monkhorst-Pack', None, None, None),
            # `center` is `'Gamma-centered'`
            (
                'Gamma-centered',
                [2, 2, 2],
                [[0.0, 1.0, 0.0, 1.0, 0.0, 1.0]],  # ! this result is weird @ndaelman-hu
                [0.0, 0.0, 0.0],
            ),
            # `center` is `'Monkhorst-Pack'`
            (
                'Monkhorst-Pack',
                [2, 2, 2],
                [
                    [-0.25, -0.25, -0.25],
                    [-0.25, -0.25, 0.25],
                    [-0.25, 0.25, -0.25],
                    [-0.25, 0.25, 0.25],
                    [0.25, -0.25, -0.25],
                    [0.25, -0.25, 0.25],
                    [0.25, 0.25, -0.25],
                    [0.25, 0.25, 0.25],
                ],
                [0.0, 0.0, 0.0],
            ),
            # Invalid `grid`
            ('Monkhorst-Pack', [-2, 2, 2], None, None),
        ],
    )
    def test_resolve_points_and_offset(
        self, center, grid, result_points, result_offset
    ):
        """
        Test the `resolve_points_and_offset` method.
        """
        k_mesh = KMesh(center=center)
        if grid is not None:
            k_mesh.grid = grid
        points, offset = k_mesh.resolve_points_and_offset(logger=logger)
        if points is not None:
            assert np.allclose(points, result_points)
        else:
            assert points == result_points
        if offset is not None:
            assert np.allclose(offset, result_offset)
        else:
            assert offset == result_offset

    @pytest.mark.parametrize(
        'system_type, is_representative, grid, reciprocal_lattice_vectors, result_get_k_line_density, result_k_line_density',
        [
            # No `grid` and `reciprocal_lattice_vectors`
            ('bulk', False, None, None, None, None),
            # No `reciprocal_lattice_vectors`
            ('bulk', False, [6, 6, 6], None, None, None),
            # No `grid`
            ('bulk', False, None, [[1, 0, 0], [0, 1, 0], [0, 0, 1]], None, None),
            # `is_representative` set to False
            (
                'bulk',
                False,
                [6, 6, 6],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                0.954929658,
                None,
            ),
            # `system_type` is not 'bulk'
            (
                'atom',
                True,
                [6, 6, 6],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                0.954929658,
                None,
            ),
            # All parameters are set
            (
                'bulk',
                True,
                [6, 6, 6],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                0.954929658,
                0.954929658,
            ),
        ],
    )
    def test_resolve_k_line_density(
        self,
        system_type: Optional[str],
        is_representative: bool,
        grid: Optional[List[int]],
        reciprocal_lattice_vectors: Optional[List[List[float]]],
        result_get_k_line_density: Optional[float],
        result_k_line_density: Optional[float],
    ):
        """
        Test the `resolve_k_line_density` and `get_k_line_density` methods
        """
        simulation = generate_k_space_simulation(
            system_type=system_type,
            is_representative=is_representative,
            reciprocal_lattice_vectors=reciprocal_lattice_vectors,
            grid=grid,
        )
        k_space = simulation.model_method[0].numerical_settings[0]
        reciprocal_lattice_vectors = k_space.reciprocal_lattice_vectors
        k_mesh = k_space.k_mesh[0]
        model_systems = simulation.model_system
        # Applying method `get_k_line_density`
        get_k_line_density_value = k_mesh.get_k_line_density(
            reciprocal_lattice_vectors=reciprocal_lattice_vectors, logger=logger
        )
        if get_k_line_density_value is not None:
            assert np.isclose(
                get_k_line_density_value.to('angstrom').magnitude,
                result_get_k_line_density,
            )
        else:
            assert get_k_line_density_value == result_get_k_line_density
        # Applying method `resolve_k_line_density`
        k_line_density = k_mesh.resolve_k_line_density(
            model_systems=model_systems,
            reciprocal_lattice_vectors=reciprocal_lattice_vectors,
            logger=logger,
        )
        if k_line_density is not None:
            assert np.isclose(
                k_line_density.to('angstrom').magnitude, result_k_line_density
            )
        else:
            assert k_line_density == result_k_line_density


class TestKLinePath:
    """
    Test the `KLinePath` class defined in `numerical_settings.py`.
    """

    @pytest.mark.parametrize(
        'high_symmetry_path_names, high_symmetry_path_values, result',
        [
            (None, None, False),
            ([], [], False),
            (['Gamma', 'X', 'Y'], None, False),
            ([], [[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0]], False),
            (['Gamma', 'X', 'Y'], [[0, 0, 0], [0.5, 0, 0], [0, 0.5, 0]], True),
        ],
    )
    def test_validate_high_symmetry_path(
        self,
        high_symmetry_path_names: List[str],
        high_symmetry_path_values: List[List[float]],
        result: bool,
    ):
        """
        Test the `validate_high_symmetry_path` method.
        """
        k_line_path = generate_k_line_path(
            high_symmetry_path_names=high_symmetry_path_names,
            high_symmetry_path_values=high_symmetry_path_values,
        )
        assert k_line_path.validate_high_symmetry_path(logger=logger) == result

    @pytest.mark.parametrize(
        'reciprocal_lattice_vectors, high_symmetry_path_names, result',
        [
            (None, None, []),
            ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], None, []),
            (
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                ['Gamma', 'X', 'R'],
                [[0, 0, 0], [0, 0.5, 0], [0.5, 0.5, 0.5]],
            ),
        ],
    )
    def test_resolve_high_symmetry_path_values(
        self,
        reciprocal_lattice_vectors: Optional[List[List[float]]],
        high_symmetry_path_names: List[str],
        result: List[float],
    ):
        """
        Test the `resolve_high_symmetry_path_values` method. Only testing the valid situation in which the `ModelSystem` normalization worked.
        """
        # `ModelSystem.normalize()` need to extract `bulk` as a type.
        simulation = generate_k_space_simulation(
            pbc=[True, True, True],
            reciprocal_lattice_vectors=reciprocal_lattice_vectors,
            high_symmetry_path_names=high_symmetry_path_names,
            high_symmetry_path_values=None,
        )
        model_system = simulation.model_system[0]
        model_system.normalize(EntryArchive(), logger)  # normalize to extract symmetry

        # `KLinePath` can be understood as a `KMeshBase` section
        k_line_path = simulation.model_method[0].numerical_settings[0].k_line_path
        high_symmetry_points_values = k_line_path.resolve_high_symmetry_path_values(
            simulation.model_system, reciprocal_lattice_vectors, logger
        )
        assert high_symmetry_points_values == result

    def test_get_high_symmetry_path_norm(self, k_line_path: KLinePath):
        """
        Test the `get_high_symmetry_path_norm` method.
        """
        rlv = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]) * ureg('1/meter')
        high_symmetry_path_norms = k_line_path.get_high_symmetry_path_norms(
            reciprocal_lattice_vectors=rlv, logger=logger
        )
        hs_points = [0, 0.5, 0.5 + 1 / np.sqrt(2), 1 + 1 / np.sqrt(2)]
        for i, val in enumerate(hs_points):
            assert np.isclose(high_symmetry_path_norms[i].magnitude, val)

    def test_resolve_points(self, k_line_path: KLinePath):
        """
        Test the `resolve_points` method.
        """
        rlv = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]) * ureg('1/meter')
        hs_points = [0, 0.5, 0.5 + 1 / np.sqrt(2), 1 + 1 / np.sqrt(2)]
        # Define paths
        gamma_x = np.linspace(hs_points[0], hs_points[1], num=5)
        x_y = np.linspace(hs_points[1], hs_points[2], num=5)
        y_gamma = np.linspace(hs_points[2], hs_points[3], num=5)
        points_norm = np.concatenate((gamma_x, x_y, y_gamma))
        k_line_path.resolve_points(
            points_norm=points_norm, reciprocal_lattice_vectors=rlv, logger=logger
        )
        assert len(points_norm) == len(k_line_path.points)
        points = np.array(
            [
                [0.0, 0.0, 0.0],  # 'Gamma'
                [0.125, 0.0, 0.0],
                [0.25, 0.0, 0.0],
                [0.375, 0.0, 0.0],
                [0.5, 0.0, 0.0],  # 'X'
                [0.4, 0.1, 0.0],
                [0.3, 0.2, 0.0],
                [0.2, 0.3, 0.0],
                [0.1, 0.4, 0.0],
                [0.0, 0.5, 0.0],  # 'Y'
                [0.0, 0.4, 0.0],
                [0.0, 0.3, 0.0],
                [0.0, 0.2, 0.0],
                [0.0, 0.1, 0.0],
                [0.0, 0.0, 0.0],  # 'Gamma'
            ]
        )
        assert np.allclose(k_line_path.points, points)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, Optional

import numpy as np

from nomad.datamodel.data import ArchiveSection
from nomad.metainfo import Quantity

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.numerical_settings import (
    KLinePath as KLinePathSettings,
    KMesh as KMeshSettings,
)


class Variables(ArchiveSection):
    """
    Variables over which the physical property varies, and they are defined as grid points, i.e., discretized
    values by `n_points` and `points`. These are used to calculate the `shape` of the physical property.
    """

    name = Quantity(
        type=str,
        default='Custom',
        description="""
        Name of the variable.
        """,
    )

    n_points = Quantity(
        type=int,
        description="""
        Number of points in which the variable is discretized.
        """,
    )

    points = Quantity(
        type=np.float64,
        # shape=['n_points'],  # ! if defined, this breaks using `points` as refs (e.g., `KMesh.points`)
        description="""
        Points in which the variable is discretized. It might be overwritten with specific units.
        """,
    )

    # ? Do we need to add `points_error`?

    def get_n_points(self, logger: 'BoundLogger') -> Optional[int]:
        """
        Get the number of grid points from the `points` list. If `n_points` is previously defined
        and does not coincide with the length of `points`, a warning is issued and this function re-assigns `n_points`
        as the length of `points`.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[int]): The number of points.
        """
        if self.points is not None and len(self.points) > 0:
            if self.n_points != len(self.points) and self.n_points is not None:
                logger.warning(
                    f'The stored `n_points`, {self.n_points}, does not coincide with the length of `points`, '
                    f'{len(self.points)}. We will re-assign `n_points` as the length of `points`.'
                )
            return len(self.points)
        return self.n_points

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Setting `n_points` if these are not defined
        self.n_points = self.get_n_points(logger)


class Temperature(Variables):
    """ """

    points = Quantity(
        type=np.float64,
        unit='kelvin',
        shape=['n_points'],
        description="""
        Points in which the temperature is discretized.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


# ! This needs to be fixed as it gives errors when running normalizers with conflicting names (ask Area D)
class Energy2(Variables):
    """ """

    points = Quantity(
        type=np.float64,
        unit='joule',
        shape=['n_points'],
        description="""
        Points in which the energy is discretized.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class WignerSeitz(Variables):
    """
    Wigner-Seitz points in which the real space is discretized. This variable is used to define `HoppingMatrix(PhysicalProperty)` and
    other inter-cell properties. See, e.g., https://en.wikipedia.org/wiki/Wigner–Seitz_cell.
    """

    points = Quantity(
        type=np.float64,
        shape=['n_points', 3],
        description="""
        Wigner-Seitz points with respect to the origin cell, (0, 0, 0). These are 3D arrays stored in fractional coordinates.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Frequency(Variables):
    """ """

    points = Quantity(
        type=np.float64,
        unit='joule',
        shape=['n_points'],
        description="""
        Points in which the frequency is discretized in joules.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class KMesh(Variables):
    """
    K-point mesh over which the physical property is calculated. This is used to define `ElectronicEigenvalues(PhysicalProperty)` and
    other k-space properties. The `points` are obtained from a reference to the `NumericalSettings` section, `KMesh(NumericalSettings)`.
    """

    points = Quantity(
        type=KMeshSettings.points,
        description="""
        Reference to the `KMesh.points` over which the physical property is calculated. These are 3D arrays stored in fractional coordinates.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class KLinePath(Variables):
    """ """

    points = Quantity(
        type=KLinePathSettings.points,
        description="""
        Reference to the `KLinePath.points` in which the physical property is calculated. These are 3D arrays stored in fractional coordinates.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from typing import Optional, List

from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.model_system import ModelSystem
from nomad_simulations.schema_packages.numerical_settings import SelfConsistency
from nomad_simulations.schema_packages.outputs import Outputs, SCFOutputs
from nomad_simulations.schema_packages.properties import ElectronicBandGap

from . import logger
from .conftest import generate_simulation, generate_scf_electronic_band_gap_template


class TestOutputs:
    """
    Test the `Outputs` class defined in `outputs.py`.
    """

    def test_number_of_properties(self):
        """
        Test how many properties are defined under `Outputs` and its order. This test is done in order to control better
        which properties are already defined and in which order to control their normalizations
        """
        outputs = Outputs()
        assert len(outputs.m_def.all_sub_sections) == 12
        defined_properties = [
            'fermi_levels',
            'chemical_potentials',
            'crystal_field_splittings',
            'hopping_matrices',
            'electronic_eigenvalues',
            'electronic_band_gaps',
            'electronic_dos',
            'fermi_surfaces',
            'electronic_band_structures',
            'permittivities',
            'absorption_spectra',
            'xas_spectra',
        ]
        assert list(outputs.m_def.all_sub_sections.keys()) == defined_properties

    @pytest.mark.parametrize(
        'band_gaps, values, result_length, result',
        [
            # no properties to extract
            ([], [], 0, []),
            # non-spin polarized case
            ([ElectronicBandGap(variables=[])], [2.0], 0, []),
            # spin polarized case
            (
                [
                    ElectronicBandGap(variables=[], spin_channel=0),
                    ElectronicBandGap(variables=[], spin_channel=1),
                ],
                [1.0, 1.5],
                2,
                [
                    ElectronicBandGap(variables=[], spin_channel=0),
                    ElectronicBandGap(variables=[], spin_channel=1),
                ],
            ),
        ],
    )
    def test_extract_spin_polarized_properties(
        self,
        band_gaps: List[ElectronicBandGap],
        values: List[float],
        result_length: int,
        result: List[ElectronicBandGap],
    ):
        """
        Test the `extract_spin_polarized_property` method.

        Args:
            band_gaps (List[ElectronicBandGap]): The `ElectronicBandGap` sections to be stored under `Outputs`.
            values (List[float]): The values to be assigned to the `ElectronicBandGap` sections.
            result_length (int): The expected length extracted from `extract_spin_polarized_property`.
            result (List[ElectronicBandGap]): The expected result of the `extract_spin_polarized_property` method.
        """
        outputs = Outputs()

        for i, band_gap in enumerate(band_gaps):
            band_gap.value = values[i]
            outputs.electronic_band_gaps.append(band_gap)
        gaps = outputs.extract_spin_polarized_property(
            property_name='electronic_band_gaps'
        )
        assert len(gaps) == result_length
        if len(result) > 0:
            for i, result_gap in enumerate(result):
                result_gap.value = values[i]
                # ? comparing the sections does not work
                assert gaps[i].value == result_gap.value
        else:
            assert gaps == result

    @pytest.mark.parametrize(
        'model_system',
        [(None), (ModelSystem(name='example'))],
    )
    def test_set_model_system_ref(self, model_system: Optional[ModelSystem]):
        """
        Test the `set_model_system_ref` method.

        Args:
            model_system (Optional[ModelSystem]): The `ModelSystem` to be tested for the `model_system_ref` reference
            stored in `Outputs`.
        """
        outputs = Outputs()
        simulation = generate_simulation(model_system=model_system, outputs=outputs)
        model_system_ref = outputs.set_model_system_ref()
        if model_system is not None:
            assert model_system_ref == simulation.model_system[-1]
            assert model_system_ref.name == 'example'
        else:
            assert model_system_ref is None

    @pytest.mark.parametrize(
        'model_system',
        [(None), (ModelSystem(name='example'))],
    )
    def test_normalize(self, model_system: Optional[ModelSystem]):
        """
        Test the `normalize` method.

        Args:
            model_system (Optional[ModelSystem]): The expected `model_system_ref` obtained after normalization and
            initially stored under `Simulation.model_system[0]`.
        """
        outputs = Outputs()
        simulation = generate_simulation(model_system=model_system, outputs=outputs)
        outputs.normalize(archive=EntryArchive(), logger=logger)
        if model_system is not None:
            assert outputs.model_system_ref == simulation.model_system[-1]
            assert outputs.model_system_ref.name == 'example'
        else:
            assert outputs.model_system_ref is None


class TestSCFOutputs:
    """
    Test the `SCFOutputs` class defined in `outputs.py`.
    """

    @pytest.mark.parametrize(
        'scf_last_steps, i_property, values, scf_parameters, result',
        [
            # empty `scf_last_steps`
            ([], 0, None, None, []),
            # length of `scf_last_steps` is different from 2
            ([Outputs()], 0, None, None, []),
            # no property matching `'electronic_band_gaps'` stored under the `scf_last_steps`
            ([Outputs(), Outputs()], 0, None, None, []),
            # `i_property` is out of range
            (
                [
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                ],
                2,
                None,
                None,
                [],
            ),
            # no `value` stored in the `scf_last_steps` property
            (
                [
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                ],
                0,
                None,
                None,
                [],
            ),
            # no `SelfConsistency` section and `threshold_change_unit` defined and macthing units for property `value` (`'joule'`)
            (
                [
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                ],
                0,
                [1.0, 2.0],
                None,
                [],
            ),
            # valid case
            (
                [
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                    Outputs(electronic_band_gaps=[ElectronicBandGap()]),
                ],
                0,
                [1.0, 2.0],
                SelfConsistency(threshold_change=1e-6, threshold_change_unit='joule'),
                [1.0, 2.0],
            ),
        ],
    )
    def test_get_last_scf_steps_value(
        self,
        scf_last_steps: List[Outputs],
        i_property: int,
        values: List[float],
        scf_parameters: Optional[SelfConsistency],
        result: List[float],
    ):
        scf_outputs = SCFOutputs()
        for i, scf_step in enumerate(scf_last_steps):
            property_section = getattr(scf_step, 'electronic_band_gaps')
            if property_section is not None and values is not None:
                property_section[i_property].value = values[i]
        scf_values = scf_outputs.get_last_scf_steps_value(
            scf_last_steps=scf_last_steps,
            property_name='electronic_band_gaps',
            i_property=i_property,
            scf_parameters=scf_parameters,
            logger=logger,
        )
        assert scf_values == result

    @pytest.mark.parametrize(
        'n_scf_steps, threshold_change, property_name, i_property, result',
        [
            # `n_scf_steps` is less than 2
            (0, None, '', 0, False),
            # no `self_consistency_ref` section
            (5, None, '', 0, False),
            # no property matching `property_name`
            (5, None, 'fermi_levels', 0, False),
            # `i_property` is out of range
            (5, None, 'electronic_band_gaps', 2, False),
            # property is not converged
            (5, 1e-5, 'electronic_band_gaps', 0, False),
            # valid case: property is converged
            (5, 1e-3, 'electronic_band_gaps', 0, True),
        ],
    )
    def test_resolve_is_scf_converged(
        self,
        n_scf_steps: int,
        threshold_change: Optional[float],
        property_name: str,
        i_property: int,
        result: bool,
    ):
        """
        Test the `resolve_is_scf_converged` method.

        Args:
            n_scf_steps (int): The number of SCF steps to add under `SCFOutputs`.
            threshold_change (Optional[float]): The threshold change to be used for the SCF convergence.
            property_name (str): The name of the property to be tested for convergence.
            i_property (int): The index of the property to be tested for convergence.
            result (bool): The expected result of the `resolve_is_scf_converged` method.
        """
        scf_outputs = generate_scf_electronic_band_gap_template(
            n_scf_steps=n_scf_steps, threshold_change=threshold_change
        )
        is_scf_converged = scf_outputs.resolve_is_scf_converged(
            property_name=property_name,
            i_property=i_property,
            physical_property=scf_outputs.electronic_band_gaps[0],
            logger=logger,
        )
        assert is_scf_converged == result

    @pytest.mark.parametrize(
        'n_scf_steps, threshold_change, result',
        [
            # `n_scf_steps` is less than 2
            (0, None, False),
            # no `self_consistency_ref` section
            (5, None, False),
            # property is not converged
            (5, 1e-5, False),
            # valid case: property is converged
            (5, 1e-3, True),
        ],
    )
    def test_normalize(
        self,
        n_scf_steps: int,
        threshold_change: float,
        result: bool,
    ):
        """
        Test the `normalize` method.

        Args:
            n_scf_steps (int): The number of SCF steps to add under `SCFOutputs`.
            threshold_change (Optional[float]): The threshold change to be used for the SCF convergence.
            result (bool): The expected result after normalization and population of the `is_scf_converged` quantity.
        """
        scf_outputs = generate_scf_electronic_band_gap_template(
            n_scf_steps=n_scf_steps, threshold_change=threshold_change
        )
        scf_outputs.normalize(EntryArchive(), logger)
        assert scf_outputs.electronic_band_gaps[0].is_scf_converged == result



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, List, Optional

from nomad.datamodel.data import ArchiveSection
from nomad.datamodel.metainfo.annotations import ELNAnnotation
from nomad.metainfo import Quantity, SubSection

if TYPE_CHECKING:
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.model_system import ModelSystem
from nomad_simulations.schema_packages.numerical_settings import SelfConsistency
from nomad_simulations.schema_packages.physical_property import PhysicalProperty
from nomad_simulations.schema_packages.properties import (
    AbsorptionSpectrum,
    ChemicalPotential,
    CrystalFieldSplitting,
    ElectronicBandGap,
    ElectronicBandStructure,
    ElectronicDensityOfStates,
    ElectronicEigenvalues,
    FermiLevel,
    FermiSurface,
    HoppingMatrix,
    Permittivity,
    XASSpectrum,
)


class Outputs(ArchiveSection):
    """
    Output properties of a simulation. This base class can be used for inheritance in any of the output properties
    defined in this schema.

    It contains references to the specific sections used to obtain the output properties, as well as
    information if the output `is_derived` from another output section or directly parsed from the simulation output files.
    """

    # TODO add time quantities

    normalizer_level = 2

    model_system_ref = Quantity(
        type=ModelSystem,
        description="""
        Reference to the `ModelSystem` section to which the output property references to and on
        on which the simulation is performed.
        """,
        a_eln=ELNAnnotation(component='ReferenceEditQuantity'),
    )

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # List of properties
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    fermi_levels = SubSection(sub_section=FermiLevel.m_def, repeats=True)

    chemical_potentials = SubSection(sub_section=ChemicalPotential.m_def, repeats=True)

    crystal_field_splittings = SubSection(
        sub_section=CrystalFieldSplitting.m_def, repeats=True
    )

    hopping_matrices = SubSection(sub_section=HoppingMatrix.m_def, repeats=True)

    electronic_eigenvalues = SubSection(
        sub_section=ElectronicEigenvalues.m_def, repeats=True
    )

    electronic_band_gaps = SubSection(sub_section=ElectronicBandGap.m_def, repeats=True)

    electronic_dos = SubSection(
        sub_section=ElectronicDensityOfStates.m_def, repeats=True
    )

    fermi_surfaces = SubSection(sub_section=FermiSurface.m_def, repeats=True)

    electronic_band_structures = SubSection(
        sub_section=ElectronicBandStructure.m_def, repeats=True
    )

    permittivities = SubSection(sub_section=Permittivity.m_def, repeats=True)

    absorption_spectra = SubSection(sub_section=AbsorptionSpectrum.m_def, repeats=True)

    xas_spectra = SubSection(sub_section=XASSpectrum.m_def, repeats=True)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def extract_spin_polarized_property(
        self, property_name: str
    ) -> List[PhysicalProperty]:
        """
        Extracts the spin-polarized properties if present from the property name and returns them as a list of two elements in
        which each element refers to each `spin_channel`. If the return list is empty, it means that the simulation is not
        spin-polarized (i.e., `spin_channel` is not defined).

        Args:
            property_name (str): The name of the property to be extracted.

        Returns:
            (List[PhysicalProperty]): The list of spin-polarized properties.
        """
        spin_polarized_properties = []
        properties = getattr(self, property_name)
        for prop in properties:
            if prop.spin_channel is None:
                continue
            spin_polarized_properties.append(prop)
        return spin_polarized_properties

    def set_model_system_ref(self) -> Optional[ModelSystem]:
        """
        Set the reference to the last ModelSystem if this is not set in the output. This is only
        valid if there is only one ModelSystem in the parent section.

        Returns:
            (Optional[ModelSystem]): The reference to the last ModelSystem.
        """
        if self.m_parent is not None:
            model_systems = self.m_parent.model_system
            if model_systems is not None and len(model_systems) == 1:
                return model_systems[-1]
        return None

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Set ref to the last ModelSystem if this is not set in the output
        if self.model_system_ref is None:
            self.model_system_ref = self.set_model_system_ref()


class SCFOutputs(Outputs):
    """
    This section contains the self-consistent (SCF) steps performed to converge an output property.

    For simplicity, we contain the SCF steps of a simulation as part of the minimal workflow defined in NOMAD,
    the `SinglePoint`, i.e., we do not split each SCF step in its own entry. Thus, each `SinglePoint`
    `Simulation` entry in NOMAD contains the final output properties and all the SCF steps.
    """

    scf_steps = SubSection(
        sub_section=Outputs.m_def,
        repeats=True,
        description="""
        Self-consistent (SCF) steps performed for converging a given output property. Note that the SCF steps belong to
        the same minimal `Simulation` workflow entry which is known as `SinglePoint`.
        """,
    )

    def get_last_scf_steps_value(
        self,
        scf_last_steps: List[Outputs],
        property_name: str,
        i_property: int,
        scf_parameters: Optional[SelfConsistency],
        logger: 'BoundLogger',
    ) -> Optional[list]:
        """
        Get the last two SCF values' magnitudes of a physical property and appends then in a list.

        Args:
            scf_last_steps (List[Outputs]): The list of SCF steps. This must be of length 2 in order to the method to work.
            property_name (str): The name of the physical property.
            i_property (int): The index of the physical property.
            scf_parameters (Optional[SelfConsistency]): The self-consistency parameters section stored under `ModelMethod`.
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[list]): The list of the last two SCF values (in magnitude) of the physical property.
        """
        # Initial check
        if len(scf_last_steps) != 2:
            logger.warning(
                '`scf_last_steps` needs to be of length 2, pointing to the last 2 SCF steps performed in the simulation.'
            )
            return []

        scf_values = []
        for step in scf_last_steps:
            try:
                scf_phys_property = getattr(step, property_name)[i_property]
                if scf_phys_property.value.u != scf_parameters.threshold_change_unit:
                    logger.error(
                        f'The units of the `scf_step.{property_name}.value` does not coincide with the units of the `self_consistency_ref.threshold_unit`.'
                    )
                    return []
            except Exception:
                return []
            scf_values.append(scf_phys_property.value.magnitude)
        return scf_values

    def resolve_is_scf_converged(
        self,
        property_name: str,
        i_property: int,
        physical_property: PhysicalProperty,
        logger: 'BoundLogger',
    ) -> bool:
        """
        Resolves if the physical property is converged or not after a SCF process. This is only ran
        when there are at least two `scf_steps` elements.

        Returns:
            (bool): Boolean indicating whether the physical property is converged or not after a SCF process.
        """
        # Check that there are at least 2 `scf_steps`
        if len(self.scf_steps) < 2:
            logger.warning('The SCF normalization needs at least two SCF steps.')
            return False
        scf_last_steps = self.scf_steps[-2:]

        # Check for `self_consistency_ref` section
        scf_parameters = physical_property.self_consistency_ref
        if scf_parameters is None:
            return False

        # Extract the value.magnitude of the `physical_property` to be checked if converged or not
        scf_values = self.get_last_scf_steps_value(
            scf_last_steps=scf_last_steps,
            property_name=property_name,
            i_property=i_property,
            scf_parameters=scf_parameters,
            logger=logger,
        )
        if scf_values is None or len(scf_values) != 2:
            logger.warning(
                f'The SCF normalization could not resolve the SCF values for the property `{property_name}`.'
            )
            return False

        # Compare with the `threshold_change`
        scf_diff = abs(scf_values[0] - scf_values[1])
        threshold_change = scf_parameters.threshold_change
        if scf_diff <= threshold_change:
            return True
        else:
            logger.info(
                f'The SCF process for the property `{property_name}` did not converge.'
            )
            return False

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Resolve the `is_scf_converged` flag for all SCF obtained properties
        for property_name in self.m_def.all_sub_sections.keys():
            # Skip the `scf_steps` and `custom_physical_property` sub-sections
            if property_name == 'scf_steps':
                continue

            # Check if the physical property with that property name is populated
            phys_properties = getattr(self, property_name)
            if phys_properties is None:
                continue
            if not isinstance(phys_properties, list):
                phys_properties = [phys_properties]

            # Loop over the physical property of the same m_def type and set `is_scf_converged`
            for i_property, phys_property in enumerate(phys_properties):
                phys_property.is_scf_converged = self.resolve_is_scf_converged(
                    property_name=property_name,
                    i_property=i_property,
                    physical_property=phys_property,
                    logger=logger,
                )



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import numpy as np
from typing import Optional, List, Union

from nomad.units import ureg
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.properties import ElectronicBandGap
from nomad_simulations.schema_packages.variables import Temperature

from . import logger


class TestElectronicBandGap:
    """
    Test the `ElectronicBandGap` class defined in `properties/band_gap.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    def test_default_quantities(self):
        """
        Test the default quantities assigned when creating an instance of the `ElectronicBandGap` class.
        """
        electronic_band_gap = ElectronicBandGap()
        assert (
            electronic_band_gap.iri
            == 'http://fairmat-nfdi.eu/taxonomy/ElectronicBandGap'
        )
        assert electronic_band_gap.name == 'ElectronicBandGap'
        assert electronic_band_gap.rank == []

    @pytest.mark.parametrize(
        'value, result',
        [
            (0.0, 0.0),
            (1.0, 1.0),
            (-1.0, None),
            ([1.0, 2.0, -1.0], None),
        ],
    )
    def test_validate_values(self, value: Union[List[float], float], result: float):
        """
        Test the `validate_values` method.
        """
        if isinstance(value, list):
            electronic_band_gap = ElectronicBandGap(
                variables=[Temperature(points=[1, 2, 3] * ureg.kelvin)]
            )
        else:
            electronic_band_gap = ElectronicBandGap()
        electronic_band_gap.value = value * ureg.joule
        validated_value = electronic_band_gap.validate_values(logger)
        if validated_value is not None:
            assert np.isclose(validated_value.magnitude, result)
        else:
            assert validated_value == result

    @pytest.mark.parametrize(
        'momentum_transfer, type, result',
        [
            (None, None, None),
            (None, 'direct', 'direct'),
            (None, 'indirect', 'indirect'),
            ([], None, None),
            ([], 'direct', 'direct'),
            ([], 'indirect', 'indirect'),
            ([[0, 0, 0]], None, None),
            ([[0, 0, 0]], 'direct', None),
            ([[0, 0, 0]], 'indirect', None),
            ([[0, 0, 0], [0, 0, 0]], None, 'direct'),
            ([[0, 0, 0], [0, 0, 0]], 'direct', 'direct'),
            ([[0, 0, 0], [0, 0, 0]], 'indirect', 'direct'),
            ([[0, 0, 0], [0.5, 0.5, 0.5]], None, 'indirect'),
            ([[0, 0, 0], [0.5, 0.5, 0.5]], 'direct', 'indirect'),
            ([[0, 0, 0], [0.5, 0.5, 0.5]], 'indirect', 'indirect'),
        ],
    )
    def test_resolve_type(
        self, momentum_transfer: Optional[List[float]], type: str, result: Optional[str]
    ):
        """
        Test the `resolve_type` method.
        """
        electronic_band_gap = ElectronicBandGap(
            variables=[],
            momentum_transfer=momentum_transfer,
            type=type,
        )
        assert electronic_band_gap.resolve_type(logger) == result

    def test_normalize(self):
        """
        Test the `normalize` method for two different ElectronicBandGap instantiations, one with a scalar
        `value` and another with a temperature-dependent `value`
        """
        scalar_band_gap = ElectronicBandGap(variables=[], type='direct')
        scalar_band_gap.value = 1.0 * ureg.joule
        scalar_band_gap.normalize(EntryArchive(), logger)
        assert scalar_band_gap.type == 'direct'
        assert np.isclose(scalar_band_gap.value.magnitude, 1.0)

        t_dependent_band_gap = ElectronicBandGap(
            variables=[Temperature(points=[0, 10, 20, 30] * ureg.kelvin)],
            type='direct',
        )
        t_dependent_band_gap.value = [1.0, 2.0, 3.0, 4.0] * ureg.joule
        t_dependent_band_gap.normalize(EntryArchive(), logger)
        assert t_dependent_band_gap.type == 'direct'
        assert (
            np.isclose(t_dependent_band_gap.value.magnitude, [1.0, 2.0, 3.0, 4.0])
        ).all()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import numpy as np
import pytest
from typing import List, Optional

from nomad.units import ureg
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_system import ModelSystem, AtomicCell
from nomad_simulations.schema_packages.atoms_state import AtomsState, OrbitalsState
from nomad_simulations.schema_packages.model_method import ModelMethod
from nomad_simulations.schema_packages.numerical_settings import (
    SelfConsistency,
    KSpace,
    KMesh as KMeshSettings,
    KLinePath as KLinePathSettings,
)
from nomad_simulations.schema_packages.outputs import Outputs, SCFOutputs
from nomad_simulations.schema_packages.variables import Energy2 as Energy, KLinePath
from nomad_simulations.schema_packages.properties import (
    ElectronicBandGap,
    DOSProfile,
    ElectronicDensityOfStates,
    ElectronicEigenvalues,
)

from . import logger

if os.getenv('_PYTEST_RAISE', '0') != '0':

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


def generate_simulation(
    model_system: Optional[ModelSystem] = None,
    model_method: Optional[ModelMethod] = None,
    outputs: Optional[Outputs] = None,
) -> Simulation:
    """
    Generate a `Simulation` section with the main sub-sections, `ModelSystem`, `ModelMethod`, and `Outputs`. If `ModelSystem`
    and `Outputs` are set, then it adds `ModelSystem` as a reference in `Outputs`.
    """
    simulation = Simulation()
    if model_method is not None:
        simulation.model_method.append(model_method)
    if model_system is not None:
        simulation.model_system.append(model_system)
    if outputs is not None:
        simulation.outputs.append(outputs)
        outputs.model_system_ref = model_system
    return simulation


def generate_model_system(
    type: str = 'original',
    system_type: str = 'bulk',
    positions: List[List[float]] = [[0, 0, 0], [0.5, 0.5, 0.5]],
    lattice_vectors: List[List[float]] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    chemical_symbols: List[str] = ['Ga', 'As'],
    orbitals_symbols: List[List[str]] = [['s'], ['px', 'py']],
    is_representative: bool = True,
    pbc: List[bool] = [False, False, False],
) -> Optional[ModelSystem]:
    """
    Generate a `ModelSystem` section with the given parameters.
    """
    if len(chemical_symbols) != len(orbitals_symbols):
        return None

    model_system = ModelSystem(type=system_type, is_representative=is_representative)
    atomic_cell = AtomicCell(
        type=type,
        positions=positions * ureg.angstrom,
        lattice_vectors=lattice_vectors * ureg.angstrom,
        periodic_boundary_conditions=pbc,
    )
    model_system.cell.append(atomic_cell)

    # Add atoms_state to the model_system
    atoms_state = []
    for element, orbitals in zip(chemical_symbols, orbitals_symbols):
        orbitals_state = []
        for orbital in orbitals:
            orbitals_state.append(
                OrbitalsState(
                    l_quantum_symbol=orbital[0], ml_quantum_symbol=orbital[1:]
                )
            )  # TODO add this split setter as part of the `OrbitalsState` methods
        atom_state = AtomsState(chemical_symbol=element, orbitals_state=orbitals_state)
        # and obtain the atomic number for each AtomsState
        atom_state.normalize(EntryArchive(), logger)
        atoms_state.append(atom_state)
    atomic_cell.atoms_state = atoms_state
    return model_system


def generate_atomic_cell(
    lattice_vectors: List[List[float]] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    positions: Optional[list] = None,
    periodic_boundary_conditions: List[bool] = [False, False, False],
    chemical_symbols: List[str] = ['H', 'H', 'O'],
    atomic_numbers: List[int] = [1, 1, 8],
) -> AtomicCell:
    """
    Generate an `AtomicCell` section with the given parameters.
    """
    # Define positions if not provided
    if positions is None and chemical_symbols is not None:
        n_atoms = len(chemical_symbols)
        positions = [[i / n_atoms, i / n_atoms, i / n_atoms] for i in range(n_atoms)]

    # Define the atomic cell
    atomic_cell = AtomicCell(periodic_boundary_conditions=periodic_boundary_conditions)
    if lattice_vectors:
        atomic_cell.lattice_vectors = lattice_vectors * ureg('angstrom')
    if positions:
        atomic_cell.positions = positions * ureg('angstrom')

    # Add the elements information
    for index, atom in enumerate(chemical_symbols):
        atom_state = AtomsState()
        setattr(atom_state, 'chemical_symbol', atom)
        atomic_number = atom_state.resolve_atomic_number(logger=logger)
        assert atomic_number == atomic_numbers[index]
        atom_state.atomic_number = atomic_number
        atomic_cell.atoms_state.append(atom_state)

    return atomic_cell


def generate_scf_electronic_band_gap_template(
    n_scf_steps: int = 5,
    threshold_change: Optional[float] = 1e-3,
    threshold_change_unit: Optional[str] = 'joule',
) -> SCFOutputs:
    """
    Generate a `SCFOutputs` section with a template for the electronic_band_gap property.
    """
    scf_outputs = SCFOutputs()
    # Define a list of scf_steps with values of the total energy like [1, 1.1, 1.11, 1.111, etc],
    # such that the difference between one step and the next one decreases a factor of 10.
    value = None
    for i in range(n_scf_steps):
        value = 1 + sum([1 / (10**j) for j in range(1, i + 2)])
        scf_step = Outputs(electronic_band_gaps=[ElectronicBandGap(value=value)])
        scf_outputs.scf_steps.append(scf_step)
    # Add a SCF calculated PhysicalProperty
    if value is not None:
        scf_outputs.electronic_band_gaps.append(ElectronicBandGap(value=value))
    else:
        scf_outputs.electronic_band_gaps.append(ElectronicBandGap())
    # and a `SelfConsistency` ref section
    if threshold_change is not None:
        model_method = ModelMethod(
            numerical_settings=[
                SelfConsistency(
                    threshold_change=threshold_change,
                    threshold_change_unit=threshold_change_unit,
                )
            ]
        )
        simulation = generate_simulation(model_method=model_method, outputs=scf_outputs)
        scf_outputs.electronic_band_gaps[
            0
        ].self_consistency_ref = simulation.model_method[0].numerical_settings[0]
    return scf_outputs


def generate_simulation_electronic_dos(
    energy_points: List[int] = [-3, -2, -1, 0, 1, 2, 3],
) -> Simulation:
    """
    Generate a `Simulation` section with an `ElectronicDensityOfStates` section under `Outputs`. It uses
    the template of the model_system created with the `generate_model_system` function.
    """
    # Create the `Simulation` section to make refs work
    model_system = generate_model_system()
    outputs = Outputs()
    simulation = generate_simulation(model_system=model_system, outputs=outputs)

    # Populating the `ElectronicDensityOfStates` section
    variables_energy = [Energy(points=energy_points * ureg.joule)]
    electronic_dos = ElectronicDensityOfStates(variables=variables_energy)
    outputs.electronic_dos.append(electronic_dos)
    # electronic_dos.value = total_dos * ureg('1/joule')
    orbital_s_Ga_pdos = DOSProfile(
        variables=variables_energy,
        entity_ref=model_system.cell[0].atoms_state[0].orbitals_state[0],
    )
    orbital_px_As_pdos = DOSProfile(
        variables=variables_energy,
        entity_ref=model_system.cell[0].atoms_state[1].orbitals_state[0],
    )
    orbital_py_As_pdos = DOSProfile(
        variables=variables_energy,
        entity_ref=model_system.cell[0].atoms_state[1].orbitals_state[1],
    )
    orbital_s_Ga_pdos.value = [0.2, 0.5, 0, 0, 0, 0.0, 0.0] * ureg('1/joule')
    orbital_px_As_pdos.value = [1.0, 0.2, 0, 0, 0, 0.3, 0.0] * ureg('1/joule')
    orbital_py_As_pdos.value = [0.3, 0.5, 0, 0, 0, 0.5, 1.3] * ureg('1/joule')
    electronic_dos.projected_dos = [
        orbital_s_Ga_pdos,
        orbital_px_As_pdos,
        orbital_py_As_pdos,
    ]
    return simulation


def generate_k_line_path(
    high_symmetry_path_names: List[str] = ['Gamma', 'X', 'Y', 'Gamma'],
    high_symmetry_path_values: List[List[float]] = [
        [0, 0, 0],
        [0.5, 0, 0],
        [0, 0.5, 0],
        [0, 0, 0],
    ],
) -> KLinePathSettings:
    return KLinePathSettings(
        high_symmetry_path_names=high_symmetry_path_names,
        high_symmetry_path_values=high_symmetry_path_values,
    )


def generate_k_space_simulation(
    system_type: str = 'bulk',
    is_representative: bool = True,
    positions: List[List[float]] = [[0, 0, 0], [0.5, 0.5, 0.5]],
    lattice_vectors: List[List[float]] = [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    chemical_symbols: List[str] = ['Ga', 'As'],
    orbitals_symbols: List[List[str]] = [['s'], ['px', 'py']],
    pbc: List[bool] = [False, False, False],
    reciprocal_lattice_vectors: Optional[List[List[float]]] = [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ],
    high_symmetry_path_names: List[str] = ['Gamma', 'X', 'Y', 'Gamma'],
    high_symmetry_path_values: List[List[float]] = [
        [0, 0, 0],
        [0.5, 0, 0],
        [0, 0.5, 0],
        [0, 0, 0],
    ],
    klinepath_points: Optional[List[float]] = None,
    grid=[6, 6, 6],
) -> Simulation:
    model_system = generate_model_system(
        system_type=system_type,
        is_representative=is_representative,
        positions=positions,
        lattice_vectors=lattice_vectors,
        chemical_symbols=chemical_symbols,
        orbitals_symbols=orbitals_symbols,
        pbc=pbc,
    )
    k_space = KSpace()
    # adding `reciprocal_lattice_vectors`
    if reciprocal_lattice_vectors is not None:
        k_space.reciprocal_lattice_vectors = (
            2 * np.pi * np.array(reciprocal_lattice_vectors) / ureg.angstrom
        )
    # adding `KMeshSettings
    k_mesh = KMeshSettings(grid=grid)
    k_space.k_mesh.append(k_mesh)
    # adding `KLinePathSettings`
    k_line_path = KLinePathSettings(
        high_symmetry_path_names=high_symmetry_path_names,
        high_symmetry_path_values=high_symmetry_path_values,
    )
    if klinepath_points is not None:
        k_line_path.points = klinepath_points
    k_space.k_line_path = k_line_path
    # appending `KSpace` to `ModelMethod.numerical_settings`
    model_method = ModelMethod()
    model_method.numerical_settings.append(k_space)
    return generate_simulation(model_method=model_method, model_system=model_system)


def generate_electronic_eigenvalues(
    reciprocal_lattice_vectors: Optional[List[List[float]]] = [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ],
    value: Optional[list] = [
        [3, -2],
        [3, 1],
        [4, -2],
        [5, -1],
        [4, 0],
        [2, 0],
        [2, 1],
        [4, -3],
    ],
    occupation: Optional[list] = [
        [0, 2],
        [0, 1],
        [0, 2],
        [0, 2],
        [0, 1.5],
        [0, 1.5],
        [0, 1],
        [0, 2],
    ],
    highest_occupied: Optional[float] = None,
    lowest_unoccupied: Optional[float] = None,
) -> ElectronicEigenvalues:
    """
    Generate an `ElectronicEigenvalues` section with the given parameters.
    """
    outputs = Outputs()
    k_space = KSpace(
        k_line_path=KLinePathSettings(
            points=[
                [0, 0, 0],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 1, 0],
                [1, 0, 1],
                [0, 1, 1],
                [1, 1, 1],
            ]
        )
    )
    model_method = ModelMethod(numerical_settings=[k_space])
    if reciprocal_lattice_vectors is not None and len(reciprocal_lattice_vectors) > 0:
        k_space.reciprocal_lattice_vectors = reciprocal_lattice_vectors
    simulation = generate_simulation(
        model_system=generate_model_system(),
        model_method=model_method,
        outputs=outputs,
    )
    electronic_eigenvalues = ElectronicEigenvalues(n_bands=2)
    outputs.electronic_eigenvalues.append(electronic_eigenvalues)
    electronic_eigenvalues.variables = [
        KLinePath(points=model_method.numerical_settings[0].k_line_path)
    ]
    if value is not None:
        electronic_eigenvalues.value = value
    if occupation is not None:
        electronic_eigenvalues.occupation = occupation
    electronic_eigenvalues.highest_occupied = highest_occupied
    electronic_eigenvalues.lowest_unoccupied = lowest_unoccupied
    return electronic_eigenvalues


@pytest.fixture(scope='session')
def model_system() -> ModelSystem:
    return generate_model_system()


@pytest.fixture(scope='session')
def atomic_cell() -> AtomicCell:
    return generate_atomic_cell()


@pytest.fixture(scope='session')
def scf_electronic_band_gap() -> SCFOutputs:
    return generate_scf_electronic_band_gap_template()


@pytest.fixture(scope='session')
def simulation_electronic_dos() -> Simulation:
    return generate_simulation_electronic_dos()


@pytest.fixture(scope='session')
def k_line_path() -> KLinePathSettings:
    return generate_k_line_path()


@pytest.fixture(scope='session')
def k_space_simulation() -> Simulation:
    return generate_k_space_simulation()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import pytest
from typing import Union, Optional

from nomad.units import ureg
from nomad.datamodel import EntryArchive
from nomad.metainfo import Quantity

from nomad_simulations.schema_packages.variables import Variables
from nomad_simulations.schema_packages.physical_property import (
    PhysicalProperty,
    validate_quantity_wrt_value,
)

from . import logger


class DummyPhysicalProperty(PhysicalProperty):
    value = Quantity(
        type=np.float64,
        unit='eV',
        description="""
        This value is defined in order to test the `__setattr__` method in `PhysicalProperty`.
        """,
    )


class TestPhysicalProperty:
    """
    Test the `PhysicalProperty` class defined in `physical_property.py`.
    """

    @pytest.mark.parametrize(
        'rank, variables, result_variables_shape, result_full_shape',
        [
            ([], [], [], []),
            ([3], [], [], [3]),
            ([3, 3], [], [], [3, 3]),
            ([], [Variables(n_points=4)], [4], [4]),
            ([3], [Variables(n_points=4)], [4], [4, 3]),
            ([3, 3], [Variables(n_points=4)], [4], [4, 3, 3]),
            (
                [],
                [Variables(n_points=4), Variables(n_points=10)],
                [4, 10],
                [4, 10],
            ),
            (
                [3],
                [Variables(n_points=4), Variables(n_points=10)],
                [4, 10],
                [4, 10, 3],
            ),
            (
                [3, 3],
                [Variables(n_points=4), Variables(n_points=10)],
                [4, 10],
                [4, 10, 3, 3],
            ),
        ],
    )
    def test_static_properties(
        self,
        rank: list,
        variables: list,
        result_variables_shape: list,
        result_full_shape: list,
    ):
        """
        Test the static properties of the `PhysicalProperty` class, `variables_shape` and `full_shape`.
        """
        physical_property = PhysicalProperty(
            source='simulation',
            rank=rank,
            variables=variables,
        )
        assert physical_property.variables_shape == result_variables_shape
        assert physical_property.full_shape == result_full_shape

    def test_setattr_value(self):
        """
        Test the `__setattr__` method when setting the `value` quantity of a physical property.
        """
        physical_property = DummyPhysicalProperty(
            source='simulation',
            rank=[3, 3],
            variables=[Variables(n_points=4), Variables(n_points=10)],
        )
        # `physical_property.value` must have full_shape=[4, 10, 3, 3]
        value = np.ones((4, 10, 3, 3)) * ureg.eV
        assert physical_property.full_shape == list(value.shape)
        physical_property.value = value
        assert np.all(physical_property.value == value)

    def test_setattr_value_wrong_shape(self):
        """
        Test the `__setattr__` method when the `value` has a wrong shape.
        """
        physical_property = PhysicalProperty(
            source='simulation',
            rank=[],
            variables=[],
        )
        # `physical_property.value` must have shape=[]
        value = np.ones((3, 3))
        wrong_shape = list(value.shape)
        with pytest.raises(ValueError) as exc_info:
            physical_property.value = value
        assert (
            str(exc_info.value)
            == f'The shape of the stored `value` {wrong_shape} does not match the full shape {physical_property.full_shape} extracted from the variables `n_points` and the `shape` defined in `PhysicalProperty`.'
        )

    def test_setattr_none(self):
        """
        Test the `__setattr__` method when setting the `value` to `None`.
        """
        physical_property = PhysicalProperty(
            source='simulation',
            rank=[],
            variables=[],
        )
        with pytest.raises(ValueError) as exc_info:
            physical_property.value = None
        assert (
            str(exc_info.value)
            == f'The value of the physical property {physical_property.name} is None. Please provide a finite valid value.'
        )

    def test_is_derived(self):
        """
        Test the `normalize` and `_is_derived` methods.
        """
        # Testing a directly parsed physical property
        not_derived_physical_property = PhysicalProperty(source='simulation')
        assert not_derived_physical_property._is_derived() is False
        not_derived_physical_property.normalize(EntryArchive(), logger)
        assert not_derived_physical_property.is_derived is False
        # Testing a derived physical property
        derived_physical_property = PhysicalProperty(
            source='analysis',
            physical_property_ref=not_derived_physical_property,
        )
        assert derived_physical_property._is_derived() is True
        derived_physical_property.normalize(EntryArchive(), logger)
        assert derived_physical_property.is_derived is True


# testing `validate_quantity_wrt_value` decorator
class ValidatingClass:
    def __init__(self, value=None, occupation=None):
        self.value = value
        self.occupation = occupation

    @validate_quantity_wrt_value('occupation')
    def validate_occupation(self) -> Union[bool, np.ndarray]:
        return self.occupation


@pytest.mark.parametrize(
    'value, occupation, result',
    [
        (None, None, False),  # Both value and occupation are None
        (np.array([[1, 2], [3, 4]]), None, False),  # occupation is None
        (None, np.array([[0.5, 1], [0, 0.5]]), False),  # value is None
        (np.array([[1, 2], [3, 4]]), np.array([]), False),  # occupation is empty
        (
            np.array([[1, 2], [3, 4]]),
            np.array([[0.5, 1]]),
            False,
        ),  # Shapes do not match
        (
            np.array([[1, 2], [3, 4]]),
            np.array([[0.5, 1], [0, 0.5]]),
            np.array([[0.5, 1], [0, 0.5]]),
        ),  # Valid case (return `occupation`)
    ],
)
def test_validate_quantity_wrt_value(
    value: Optional[np.ndarray],
    occupation: Optional[np.ndarray],
    result: Union[bool, np.ndarray],
):
    """
    Test the `validate_quantity_wrt_value` decorator.
    """
    obj = ValidatingClass(value=value, occupation=occupation)
    validation = obj.validate_occupation()
    if isinstance(validation, bool):
        assert validation == result
    else:
        assert np.allclose(validation, result)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import numpy as np
from typing import List, Optional

from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.model_system import (
    Symmetry,
    ChemicalFormula,
    ModelSystem,
)

from . import logger
from .conftest import generate_atomic_cell


class TestAtomicCell:
    """
    Test the `AtomicCell`, `Cell` and `GeometricSpace` classes defined in model_system.py
    """

    @pytest.mark.parametrize(
        'chemical_symbols, atomic_numbers, formula, lattice_vectors, positions, periodic_boundary_conditions',
        [
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                'H2O',
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [False, False, False],
            ),  # full atomic cell
            (
                [],
                [1, 1, 8],
                'H2O',
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [False, False, False],
            ),  # missing chemical_symbols
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                'H2O',
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [],
                [False, False, False],
            ),  # missing positions
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                'H2O',
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1], [2, 2, 2]],
                [False, False, False],
            ),  # chemical_symbols and positions with different lengths
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                'H2O',
                [],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [False, False, False],
            ),  # missing lattice_vectors
        ],
    )
    def test_generate_ase_atoms(
        self,
        chemical_symbols: List[str],
        atomic_numbers: List[int],
        formula: str,
        lattice_vectors: List[List[float]],
        positions: List[List[float]],
        periodic_boundary_conditions: List[bool],
    ):
        """
        Test the creation of `ase.Atoms` from `AtomicCell`.

        Args:
            chemical_symbols (List[str]): List of chemical symbols.
            atomic_numbers (List[int]): List of atomic numbers.
            formula (str): Chemical formula.
            lattice_vectors (List[List[float]]): Lattice vectors.
            positions (List[List[float]]): Atomic positions.
            periodic_boundary_conditions (List[bool]): Periodic boundary conditions.
        """
        atomic_cell = generate_atomic_cell(
            lattice_vectors=lattice_vectors,
            positions=positions,
            periodic_boundary_conditions=periodic_boundary_conditions,
            chemical_symbols=chemical_symbols,
            atomic_numbers=atomic_numbers,
        )

        # Test `to_ase_atoms` function
        ase_atoms = atomic_cell.to_ase_atoms(logger=logger)
        if not chemical_symbols or len(chemical_symbols) != len(positions):
            assert ase_atoms is None
        else:
            if lattice_vectors:
                assert (ase_atoms.cell == lattice_vectors).all()
            else:
                assert (ase_atoms.cell == [0, 0, 0]).all()
            assert (ase_atoms.positions == positions).all()
            assert (ase_atoms.pbc == periodic_boundary_conditions).all()
            assert (ase_atoms.symbols.numbers == atomic_numbers).all()
            assert ase_atoms.symbols.get_chemical_formula() == formula

    @pytest.mark.parametrize(
        'chemical_symbols, atomic_numbers, lattice_vectors, positions, vectors_results, angles_results, volume',
        [
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [1.0, 1.0, 1.0],
                [90.0, 90.0, 90.0],
                1.0,
            ),  # full atomic cell
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                [[1.2, 2.3, 0], [1.2, -2.3, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [2.59422435, 2.59422435, 1.0],
                [90.0, 90.0, 124.8943768],
                5.52,
            ),  # full atomic cell with different lattice_vectors
            (
                [],
                [1, 1, 8],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [None, None, None],
                [None, None, None],
                None,
            ),  # missing chemical_symbols
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [],
                [None, None, None],
                [None, None, None],
                None,
            ),  # missing positions
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1], [2, 2, 2]],
                [None, None, None],
                [None, None, None],
                None,
            ),  # chemical_symbols and positions with different lengths
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                [],
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [0.0, 0.0, 0.0],
                [90.0, 90.0, 90.0],
                0.0,
            ),  # missing lattice_vectors
        ],
    )
    def test_geometric_space(
        self,
        chemical_symbols: List[str],
        atomic_numbers: List[int],
        lattice_vectors: List[List[float]],
        positions: List[List[float]],
        vectors_results: List[Optional[float]],
        angles_results: List[Optional[float]],
        volume: Optional[float],
    ):
        """
        Test the `GeometricSpace` quantities normalization from `AtomicCell`.

        Args:
            chemical_symbols (List[str]): List of chemical symbols.
            atomic_numbers (List[int]): List of atomic numbers.
            lattice_vectors (List[List[float]]): Lattice vectors.
            positions (List[List[float]]): Atomic positions.
            vectors_results (List[Optional[float]]): Expected lengths of cell vectors.
            angles_results (List[Optional[float]]): Expected angles between cell vectors.
            volume (Optional[float]): Expected volume of the cell.
        """
        atomic_cell = generate_atomic_cell(
            lattice_vectors=lattice_vectors,
            positions=positions,
            chemical_symbols=chemical_symbols,
            atomic_numbers=atomic_numbers,
        )

        # Get `GeometricSpace` quantities via normalization of `AtomicCell`
        atomic_cell.normalize(EntryArchive(), logger)
        # Testing lengths of cell vectors
        for index, name in enumerate(
            ['length_vector_a', 'length_vector_b', 'length_vector_c']
        ):
            quantity = getattr(atomic_cell, name)
            if quantity is not None:
                assert np.isclose(
                    quantity.to('angstrom').magnitude,
                    vectors_results[index],
                )
            else:
                assert quantity == vectors_results[index]
        # Testing angles between cell vectors
        for index, name in enumerate(
            ['angle_vectors_b_c', 'angle_vectors_a_c', 'angle_vectors_a_b']
        ):
            quantity = getattr(atomic_cell, name)
            if quantity is not None:
                assert np.isclose(
                    quantity.to('degree').magnitude,
                    angles_results[index],
                )
            else:
                assert quantity == angles_results[index]
        # Testing volume
        if atomic_cell.volume is not None:
            assert np.isclose(atomic_cell.volume.to('angstrom^3').magnitude, volume)
        else:
            assert atomic_cell.volume == volume


class TestModelSystem:
    """
    Test the `ModelSystem`, `Symmetry` and `ChemicalFormula` classes defined in model_system.py
    """

    def test_empty_chemical_formula(self):
        """
        Test the empty `ChemicalFormula` normalization if a sibling `AtomicCell` is not provided.
        """
        chemical_formula = ChemicalFormula()
        chemical_formula.normalize(EntryArchive(), logger)
        for name in ['descriptive', 'reduced', 'iupac', 'hill', 'anonymous']:
            assert getattr(chemical_formula, name) is None

    @pytest.mark.parametrize(
        'chemical_symbols, atomic_numbers, formulas',
        [
            (
                ['H', 'H', 'O'],
                [1, 1, 8],
                ['H2O', 'H2O', 'H2O', 'H2O', 'A2B'],
            ),
            (
                ['O', 'O', 'O', 'O', 'La', 'Cu', 'Cu'],
                [8, 8, 8, 8, 57, 29, 29],
                ['LaCu2O4', 'Cu2LaO4', 'LaCu2O4', 'Cu2LaO4', 'A4B2C'],
            ),
            (
                ['O', 'La', 'As', 'Fe', 'C'],
                [8, 57, 33, 26, 6],
                ['CAsFeLaO', 'AsCFeLaO', 'LaFeCAsO', 'CAsFeLaO', 'ABCDE'],
            ),
        ],
    )
    def test_chemical_formula(
        self,
        chemical_symbols: List[str],
        atomic_numbers: List[int],
        formulas: List[str],
    ):
        """
        Test the `ChemicalFormula` normalization if a sibling `AtomicCell` is created, and thus the `Formula` class can be used.

        Args:
            chemical_symbols (List[str]): List of chemical symbols.
            atomic_numbers (List[int]): List of atomic numbers.
            formulas (List[str]): List of expected formulas.
        """
        atomic_cell = generate_atomic_cell(
            chemical_symbols=chemical_symbols, atomic_numbers=atomic_numbers
        )
        chemical_formula = ChemicalFormula()
        model_system = ModelSystem(chemical_formula=chemical_formula)
        model_system.cell.append(atomic_cell)
        chemical_formula.normalize(EntryArchive(), logger)
        for index, name in enumerate(
            ['descriptive', 'reduced', 'iupac', 'hill', 'anonymous']
        ):
            assert getattr(chemical_formula, name) == formulas[index]

    @pytest.mark.parametrize(
        'positions, pbc, system_type, dimensionality',
        [
            (
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                None,
                'molecule / cluster',
                0,
            ),
            (
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [False, False, False],
                'molecule / cluster',
                0,
            ),
            (
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [True, False, False],
                '1D',
                1,
            ),
            (
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [True, True, False],
                '2D',
                2,
            ),
            (
                [[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]],
                [True, True, True],
                'bulk',
                3,
            ),
        ],
    )
    def test_system_type_and_dimensionality(
        self,
        positions: List[List[float]],
        pbc: Optional[List[bool]],
        system_type: str,
        dimensionality: int,
    ):
        """
        Test the `ModelSystem` normalization of `type` and `dimensionality` from `AtomicCell`.

        Args:
            positions (List[List[float]]): Atomic positions.
            pbc (Optional[List[bool]]): Periodic boundary conditions.
            system_type (str): Expected system type.
            dimensionality (int): Expected dimensionality.
        """
        atomic_cell = generate_atomic_cell(
            positions=positions, periodic_boundary_conditions=pbc
        )
        ase_atoms = atomic_cell.to_ase_atoms(logger=logger)
        model_system = ModelSystem()
        model_system.cell.append(atomic_cell)
        (
            resolved_system_type,
            resolved_dimensionality,
        ) = model_system.resolve_system_type_and_dimensionality(
            ase_atoms=ase_atoms, logger=logger
        )
        assert resolved_system_type == system_type
        assert resolved_dimensionality == dimensionality

    def test_symmetry(self):
        """
        Test the `Symmetry` normalization from a sibling `AtomicCell` section.
        """
        atomic_cell = generate_atomic_cell(
            periodic_boundary_conditions=[True, True, True]
        )
        assert (
            np.isclose(
                atomic_cell.lattice_vectors.to('angstrom').magnitude,
                np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
            )
        ).all()
        symmetry = Symmetry()
        primitive, conventional = symmetry.resolve_bulk_symmetry(
            original_atomic_cell=atomic_cell, logger=logger
        )
        assert symmetry.bravais_lattice == 'hR'
        assert symmetry.hall_symbol == '-R 3 2"'
        assert symmetry.point_group_symbol == '-3m'
        assert symmetry.space_group_number == 166
        assert symmetry.space_group_symbol == 'R-3m'
        assert primitive.type == 'primitive'
        assert primitive.periodic_boundary_conditions == [False, False, False]
        assert (
            np.isclose(
                primitive.lattice_vectors.to('angstrom').magnitude,
                np.array(
                    [
                        [7.07106781e-01, 4.08248290e-01, 5.77350269e-01],
                        [-7.07106781e-01, 4.08248290e-01, 5.77350269e-01],
                        [1.08392265e-17, -8.16496581e-01, 5.77350269e-01],
                    ]
                ),
            )
        ).all()
        assert conventional.type == 'conventional'
        assert (
            np.isclose(
                conventional.lattice_vectors.to('angstrom').magnitude,
                np.array(
                    [
                        [1.41421356, 0.0, 0.0],
                        [-0.70710678, 1.22474487, 0.0],
                        [0.0, 0.0, 1.73205081],
                    ]
                ),
            )
        ).all()

    def test_no_representative(self):
        """
        Test the normalization of a `ModelSystem` is not run if it is not representative.
        """
        model_system = ModelSystem(is_representative=False)
        model_system.normalize(EntryArchive(), logger)
        assert model_system.type is None
        assert model_system.dimensionality is None

    def test_empty_atomic_cell(self):
        """
        Test the normalization of a `ModelSystem` is not run if it has no `AtomicCell` child section.
        """
        model_system = ModelSystem(is_representative=True)
        model_system.normalize(EntryArchive(), logger)
        assert model_system.type is None
        assert model_system.dimensionality is None

    def test_normalize(self):
        """
        Test the full normalization of a representative `ModelSystem`.
        """
        atomic_cell = generate_atomic_cell(
            periodic_boundary_conditions=[True, True, True]
        )
        model_system = ModelSystem(is_representative=True)
        model_system.cell.append(atomic_cell)
        model_system.normalize(EntryArchive(), logger)
        # Basic quantities assertions
        assert model_system.type == 'bulk'
        assert model_system.dimensionality == 3
        # AtomicCell
        assert len(model_system.cell) == 3
        assert model_system.cell[0].type == 'original'
        assert model_system.cell[1].type == 'primitive'
        assert model_system.cell[2].type == 'conventional'
        # Symmetry
        assert len(model_system.symmetry) == 1
        assert model_system.symmetry[0].bravais_lattice == 'hR'
        assert model_system.symmetry[0].atomic_cell_ref == model_system.cell[2]
        # ChemicalFormula
        assert model_system.chemical_formula.descriptive == 'H2O'
        # ElementalComposition
        assert len(model_system.elemental_composition) == 2
        assert model_system.elemental_composition[0].element == 'H'
        assert np.isclose(model_system.elemental_composition[0].atomic_fraction, 2 / 3)
        assert model_system.elemental_composition[1].element == 'O'
        assert np.isclose(model_system.elemental_composition[1].atomic_fraction, 1 / 3)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import wraps
from typing import TYPE_CHECKING, Any, Optional

import numpy as np

from nomad import utils
from nomad.datamodel.data import ArchiveSection
from nomad.datamodel.metainfo.basesections import Entity
from nomad.metainfo import (
    URL,
    MEnum,
    Quantity,
    Reference,
    SectionProxy,
    SubSection,
)
from nomad.metainfo.metainfo import Dimension, DirectQuantity, _placeholder_quantity

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.numerical_settings import SelfConsistency
from nomad_simulations.schema_packages.variables import Variables

# We add `logger` for the `validate_quantity_wrt_value` decorator
logger = utils.get_logger(__name__)


def validate_quantity_wrt_value(name: str = ''):
    """
    Decorator to validate the existence of a quantity and its shape with respect to the `PhysicalProperty.value`
    before calling a method. An example can be found in the module `properties/band_structure.py` for the method
    `ElectronicEigenvalues.order_eigenvalues()`.

    Args:
        name (str, optional): The name of the `quantity` to validate. Defaults to ''.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # Checks if `quantity` is defined
            quantity = getattr(self, name, None)
            if quantity is None or len(quantity) == 0:
                logger.warning(f'The quantity `{name}` is not defined.')
                return False

            # Checks if `value` exists and has the same shape as `quantity`
            value = getattr(self, 'value', None)
            if value is None:
                logger.warning('The quantity `value` is not defined.')
                return False
            if value is not None and value.shape != quantity.shape:
                logger.warning(
                    f'The shape of the quantity `{name}` does not match the shape of the `value`.'
                )
                return False

            return func(self, *args, **kwargs)

        return wrapper

    return decorator


class PhysicalProperty(ArchiveSection):
    """
    A base section used to define the physical properties obtained in a simulation, experiment, or in a post-processing
    analysis. The main quantity of the `PhysicalProperty` is `value`, whose instantiation has to be overwritten in the derived classes
    when inheriting from `PhysicalProperty`. It also contains `rank`, to define the tensor rank of the physical property, and
    `variables`, to define the variables over which the physical property varies (see variables.py). This class can also store several
    string identifiers and quantities for referencing and establishing the character of a physical property.
    """

    # TODO add `errors`
    # TODO add `smearing`

    name = Quantity(
        type=str,
        description="""
        Name of the physical property. Example: `'ElectronicBandGap'`.
        """,
    )

    iri = Quantity(
        type=URL,
        description="""
        Internationalized Resource Identifier (IRI) of the physical property defined in the FAIRmat
        taxonomy, https://fairmat-nfdi.github.io/fairmat-taxonomy/.
        """,
    )

    source = Quantity(
        type=MEnum('simulation', 'measurement', 'analysis'),
        default='simulation',
        description="""
        Source of the physical property. This quantity is related with the `Activity` performed to obtain the physical
        property. Example: an `ElectronicBandGap` can be obtained from a `'simulation'` or in a `'measurement'`.
        """,
    )

    type = Quantity(
        type=str,
        description="""
        Type categorization of the physical property. Example: an `ElectronicBandGap` can be `'direct'`
        or `'indirect'`.
        """,
        # ! add more examples in the description to improve the understanding of this quantity
    )

    label = Quantity(
        type=str,
        description="""
        Label for additional classification of the physical property. Example: an `ElectronicBandGap`
        can be labeled as `'DFT'` or `'GW'` depending on the methodology used to calculate it.
        """,
        # ! add more examples in the description to improve the understanding of this quantity
    )

    rank = DirectQuantity(
        type=Dimension,
        shape=['0..*'],
        default=[],
        name='rank',
        description="""
        Rank of the tensor describing the physical property. This quantity is stored as a Dimension:
            - scalars (tensor rank 0) have `rank=[]` (`len(rank) = 0`),
            - vectors (tensor rank 1) have `rank=[a]` (`len(rank) = 1`),
            - matrices (tensor rank 2), have `rank=[a, b]` (`len(rank) = 2`),
            - etc.
        """,
    )

    variables = SubSection(sub_section=Variables.m_def, repeats=True)

    # * `value` must be overwritten in the derived classes defining its type, unit, and description
    value: Quantity = _placeholder_quantity

    entity_ref = Quantity(
        type=Entity,
        description="""
        Reference to the entity that the physical property refers to. Examples:
            - a simulated physical property might refer to the macroscopic system or instead of a specific atom in the unit
            cell. In the first case, `outputs.model_system_ref` (see outputs.py) will point to the `ModelSystem` section,
            while in the second case, `entity_ref` will point to `AtomsState` section (see atoms_state.py).
        """,
    )

    physical_property_ref = Quantity(
        type=Reference(SectionProxy('PhysicalProperty')),
        description="""
        Reference to the `PhysicalProperty` section from which the physical property was derived. If `physical_property_ref`
        is populated, the quantity `is_derived` is set to True via normalization.
        """,
    )

    is_derived = Quantity(
        type=bool,
        default=False,
        description="""
        Flag indicating whether the physical property is derived from other physical properties. We make
        the distinction between directly parsed and derived physical properties:
            - Directly parsed: the physical property is directly parsed from the simulation output files.
            - Derived: the physical property is derived from other physical properties. No extra numerical settings
                are required to calculate the physical property.
        """,
    )

    is_scf_converged = Quantity(
        type=bool,
        description="""
        Flag indicating whether the physical property is converged or not after a SCF process. This quantity is connected
        with `SelfConsistency` defined in the `numerical_settings.py` module.
        """,
    )

    self_consistency_ref = Quantity(
        type=SelfConsistency,
        description="""
        Reference to the `SelfConsistency` section that defines the numerical settings to converge the
        physical property (see numerical_settings.py).
        """,
    )

    @property
    def variables_shape(self) -> Optional[list]:
        """
        Shape of the variables over which the physical property varies. This is extracted from
        `Variables.n_points` and appended in a list.

        Example, a physical property which varies with `Temperature` and `ElectricField` will
        return `variables_shape = [n_temperatures, n_electric_fields]`.

        Returns:
            (list): The shape of the variables over which the physical property varies.
        """
        if self.variables is not None:
            return [v.get_n_points(logger) for v in self.variables]
        return []

    @property
    def full_shape(self) -> list:
        """
        Full shape of the physical property. This quantity is calculated as a concatenation of the `variables_shape`
        and `rank`:

            `full_shape = variables_shape + rank`

        where `rank` is passed as an attribute of the `PhysicalProperty` and is related with the order of
        the tensor of `value`, and `variables_shape` is obtained from the property-decorated function `variables_shape()`
        and is related with the shapes of the `variables` over which the physical property varies.

        Example: a physical property which is a 3D vector and varies with `variables=[Temperature, ElectricField]`
        will have `rank=[3]`, `variables_shape=[n_temperatures, n_electric_fields]`, and thus
        `full_shape=[n_temperatures, n_electric_fields, 3]`.

        Returns:
            (list): The full shape of the physical property.
        """
        return self.variables_shape + self.rank

    @property
    def _new_value(self) -> Quantity:
        """
        Initialize a new `Quantity` object for the `value` quantity with the correct `shape` extracted from
        the `full_shape` attribute. This copies the main attributes from `value` (`type`, `description`, `unit`).
        It is used in the `__setattr__` method.

        Returns:
            (Quantity): The new `Quantity` object for setting the `value` quantity.
        """
        value_quantity = self.m_def.all_quantities.get('value')
        if value_quantity is None:
            return None
        return Quantity(
            type=value_quantity.type,
            unit=value_quantity.unit,  # ? this can be moved to __setattr__
            description=value_quantity.description,
        )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)

        # Checking if IRI is defined
        if self.iri is None:
            logger.warning(
                'The used property is not defined in the FAIRmat taxonomy (https://fairmat-nfdi.github.io/fairmat-taxonomy/). You can contribute there if you want to extend the list of available materials properties.'
            )

        # Checking if the quantities `n_` are defined, as this are used to calculate `rank`
        for quantity, _ in self.m_def.all_quantities.items():
            if quantity.startswith('n_') and getattr(self, quantity) is None:
                raise ValueError(
                    f'`{quantity}` is not defined during initialization of the class.'
                )

    def __setattr__(self, name: str, val: Any) -> None:
        # For the special case of `value`, its `shape` needs to be defined from `_full_shape`
        if name == 'value':
            if val is None:
                raise ValueError(
                    f'The value of the physical property {self.name} is None. Please provide a finite valid value.'
                )
            _new_value = self._new_value

            # patch for when `val` does not have units and it is passed as a list (instead of np.array)
            if isinstance(val, list):
                val = np.array(val)

            # non-scalar or scalar `val`
            try:
                value_shape = list(val.shape)
            except AttributeError:
                value_shape = []

            if value_shape != self.full_shape:
                raise ValueError(
                    f'The shape of the stored `value` {value_shape} does not match the full shape {self.full_shape} '
                    f'extracted from the variables `n_points` and the `shape` defined in `PhysicalProperty`.'
                )
            _new_value.shape = self.full_shape
            if hasattr(val, 'magnitude'):
                _new_value = val.magnitude * val.u
            else:
                _new_value = val
            return super().__setattr__(name, _new_value)
        return super().__setattr__(name, val)

    def _is_derived(self) -> bool:
        """
        Resolves if the physical property is derived or not.

        Returns:
            (bool): The flag indicating whether the physical property is derived or not.
        """
        if self.physical_property_ref is not None:
            return True
        return False

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Resolve if the physical property `is_derived` or not from another physical property.
        self.is_derived = self._is_derived()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from nomad_simulations.schema_packages.utils import (
    get_sibling_section,
    is_not_representative,
    get_variables,
)
from nomad_simulations.schema_packages.model_system import (
    ModelSystem,
    AtomicCell,
    Symmetry,
)
from nomad_simulations.schema_packages.variables import Temperature, Energy2 as Energy

from . import logger


def test_get_sibling_section():
    """
    Test the `get_sibling_section` utility function.
    """
    parent_section = ModelSystem()
    section = AtomicCell(type='original')
    parent_section.cell.append(section)
    sibling_section = Symmetry()
    parent_section.symmetry.append(sibling_section)
    assert get_sibling_section(section, '', logger) is None
    assert get_sibling_section(section, 'symmetry', logger) == sibling_section
    assert get_sibling_section(sibling_section, 'cell', logger) == section
    assert get_sibling_section(section, 'symmetry', logger, index_sibling=2) is None
    section2 = AtomicCell(type='primitive')
    parent_section.cell.append(section2)
    assert (
        get_sibling_section(sibling_section, 'cell', logger, index_sibling=0) == section
    )
    assert (
        get_sibling_section(sibling_section, 'cell', logger, index_sibling=1)
        == section2
    )


def test_is_not_representative():
    """
    Test the `is_not_representative` utility function.
    """
    assert is_not_representative(None, logger) is None
    assert is_not_representative(ModelSystem(), logger)
    assert not is_not_representative(ModelSystem(is_representative=True), logger)


# ! Missing test for RusselSandersState (but this class will probably be deprecated)


@pytest.mark.parametrize(
    'variables, result, result_length',
    [
        (None, [], 0),
        ([], [], 0),
        ([Temperature()], [], 0),
        ([Temperature(), Energy(n_points=4)], [Energy(n_points=4)], 1),
        (
            [Temperature(), Energy(n_points=2), Energy(n_points=10)],
            [Energy(n_points=2), Energy(n_points=10)],
            2,
        ),
        # TODO add testing when we have variables which inherit from another variable
    ],
)
def test_get_variables(variables: list, result: list, result_length: int):
    """
    Test the `get_variables` utility function
    """
    energies = get_variables(variables, Energy)
    assert len(energies) == result_length
    for i, energy in enumerate(energies):  # asserting energies == result does not work
        assert energy.n_points == result[i].n_points



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from typing import Optional

from nomad_simulations.schema_packages.properties import (
    HoppingMatrix,
    CrystalFieldSplitting,
)


class TestHoppingMatrix:
    """
    Test the `HoppingMatrix` class defined in `properties/hopping_matrix.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    @pytest.mark.parametrize(
        'n_orbitals, rank',
        [
            (None, None),
            (3, [3, 3]),
        ],
    )
    def test_default_quantities(self, n_orbitals: Optional[int], rank: Optional[list]):
        """
        Test the default quantities assigned when creating an instance of the `HoppingMatrix` class.
        """
        if n_orbitals is None:
            with pytest.raises(ValueError) as exc_info:
                hopping_matrix = HoppingMatrix(n_orbitals=n_orbitals)
            assert (
                str(exc_info.value)
                == f'`n_orbitals` is not defined during initialization of the class.'
            )
        else:
            hopping_matrix = HoppingMatrix(n_orbitals=n_orbitals)
            assert hopping_matrix.iri == 'http://fairmat-nfdi.eu/taxonomy/HoppingMatrix'
            assert hopping_matrix.name == 'HoppingMatrix'
            assert hopping_matrix.rank == rank


class TestCrystalFieldSplitting:
    """
    Test the `CrystalFieldSplitting` class defined in `properties/hopping_matrix.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    @pytest.mark.parametrize(
        'n_orbitals, rank',
        [
            (None, None),
            (3, [3]),
        ],
    )
    def test_default_quantities(self, n_orbitals: Optional[int], rank: Optional[list]):
        """
        Test the default quantities assigned when creating an instance of the `CrystalFieldSplitting` class.
        """
        if n_orbitals is None:
            with pytest.raises(ValueError) as exc_info:
                crystal_field = CrystalFieldSplitting(n_orbitals=n_orbitals)
            assert (
                str(exc_info.value)
                == f'`n_orbitals` is not defined during initialization of the class.'
            )
        else:
            crystal_field = CrystalFieldSplitting(n_orbitals=n_orbitals)
            assert (
                crystal_field.iri
                == 'http://fairmat-nfdi.eu/taxonomy/CrystalFieldSplitting'
            )
            assert crystal_field.name == 'CrystalFieldSplitting'
            assert crystal_field.rank == rank



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from math import factorial
from typing import List, Optional

import numpy as np
from nomad.datamodel.data import ArchiveSection
from structlog.stdlib import BoundLogger


def get_sibling_section(
    section: ArchiveSection,
    sibling_section_name: str,
    logger: BoundLogger,
    index_sibling: int = 0,
) -> Optional[ArchiveSection]:
    """
    Gets the sibling section of a section by performing a seesaw move by going to the parent
    of the section and then going down to the sibling section. This is used, e.g., to get
    the `AtomicCell` section from the `Symmetry` section and by passing through the `ModelSystem`.

    Example of the sections structure:

        parent_section
          |__ section
          |__ sibling_section


    If the sibling_section is a list, it returns the element `index_sibling` of that list. If
    the sibling_section is a single section, it returns the sibling_section itself.

    Args:
        section (ArchiveSection): The section to check for its parent and retrieve the sibling_section.
        sibling_section (str): The name of the sibling_section to retrieve from the parent.
        index_sibling (int): The index of the sibling_section to retrieve if it is a list.
        logger (BoundLogger): The logger to log messages.

    Returns:
        sibling_section (ArchiveSection): The sibling_section to be returned.
    """
    if not sibling_section_name:
        logger.warning('The sibling_section_name is empty.')
        return None
    sibling_section = section.m_xpath(f'm_parent.{sibling_section_name}', dict=False)
    # If the sibling_section is a list, return the element `index_sibling` of that list
    if isinstance(sibling_section, list):
        if index_sibling >= len(sibling_section):
            logger.warning('The index of the sibling_section is out of range.')
            return None
        return sibling_section[index_sibling]
    return sibling_section


# ? Check if this utils deserves its own file after extending it
class RussellSaundersState:
    @classmethod
    def generate_Js(cls, J1: float, J2: float, rising=True):
        J_min, J_max = sorted([abs(J1), abs(J2)])
        generator = range(
            int(J_max - J_min) + 1
        )  # works for both for fermions and bosons
        if rising:
            for jj in generator:
                yield J_min + jj
        else:
            for jj in generator:
                yield J_max - jj

    @classmethod
    def generate_MJs(cls, J, rising=True):
        generator = range(int(2 * J + 1))
        if rising:
            for m in generator:
                yield -J + m
        else:
            for m in generator:
                yield J - m

    def __init__(self, *args, **kwargs):
        self.J = kwargs.get('J')
        if self.J is None:
            raise TypeError
        self.occupation = kwargs.get('occ')
        if self.occupation is None:
            raise TypeError

    @property
    def multiplicity(self):
        return 2 * self.J + 1

    @property
    def degeneracy(self):
        return factorial(self.multiplicity) / (
            factorial(self.multiplicity - self.occupation) * factorial(self.occupation)
        )


def is_not_representative(model_system, logger: BoundLogger):
    """
    Checks if the given `ModelSystem` is not representative and logs a warning.

    Args:
        model_system (ModelSystem): The `ModelSystem` to check.
        logger (BoundLogger): The logger to log messages.

    Returns:
        (bool): True if the `ModelSystem` is not representative, False otherwise.
    """
    if model_system is None:
        logger.warning('The `ModelSystem` is empty.')
        return None
    if not model_system.is_representative:
        return True
    return False


# cannot define typing with `Variables` due to circular import issue
def get_variables(
    variables: Optional[List[ArchiveSection]], variable_cls: ArchiveSection
) -> List[ArchiveSection]:
    """
    Get the list of variables which are of type `variable_cls` and appear under `variables`.

    Args:
        variables (Optional[List[Variables]]): The list of variables to check.
        variable_cls (Variables): The class of the variables to get.

    Returns:
        (List[Variables]): The list of variables which are of type `variable_cls`.
    """
    if variables is None:
        return []
    result = []
    for var in variables:
        if isinstance(var, variable_cls):
            result.append(var)
    return result


# TODO remove function in nomad.atomutils
def get_composition(children_names: List[str]) -> str:
    """
    Generates a generalized "chemical formula" based on the provided list `children_names`,
    with the format X(m)Y(n) for children_names X and Y of quantities m and n, respectively.
    """
    children_count_tup = np.unique(children_names, return_counts=True)
    formula = ''.join([f'{name}({count})' for name, count in zip(*children_count_tup)])
    return formula if formula else None



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from typing import Optional

from nomad_simulations.schema_packages.properties import FermiSurface


class TestFermiSurface:
    """
    Test the `FermiSurface` class defined in `properties/band_structure.py`.
    """

    # ! Include this initial `test_default_quantities` method when testing your PhysicalProperty classes
    @pytest.mark.parametrize(
        'n_bands, rank',
        [
            (None, None),
            (10, [10]),
        ],
    )
    def test_default_quantities(self, n_bands: Optional[int], rank: Optional[list]):
        """
        Test the default quantities assigned when creating an instance of the `HoppingMatrix` class.
        """
        if n_bands is None:
            with pytest.raises(ValueError) as exc_info:
                fermi_surface = FermiSurface(n_bands=n_bands)
            assert (
                str(exc_info.value)
                == f'`n_bands` is not defined during initialization of the class.'
            )
        else:
            fermi_surface = FermiSurface(n_bands=n_bands)
            assert fermi_surface.iri == 'http://fairmat-nfdi.eu/taxonomy/FermiSurface'
            assert fermi_surface.name == 'FermiSurface'
            assert fermi_surface.rank == rank



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, Optional

import numpy as np
import pint

from nomad.metainfo import MEnum, Quantity

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import PhysicalProperty


class ElectronicBandGap(PhysicalProperty):
    """
    Energy difference between the highest occupied electronic state and the lowest unoccupied electronic state.
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    iri = 'http://fairmat-nfdi.eu/taxonomy/ElectronicBandGap'

    type = Quantity(
        type=MEnum('direct', 'indirect'),
        description="""
        Type categorization of the electronic band gap. This quantity is directly related with `momentum_transfer` as by
        definition, the electronic band gap is `'direct'` for zero momentum transfer (or if `momentum_transfer` is `None`) and `'indirect'`
        for finite momentum transfer.

        Note: in the case of finite `variables`, this quantity refers to all of the `value` in the array.
        """,
    )

    momentum_transfer = Quantity(
        type=np.float64,
        shape=[2, 3],
        description="""
        If the electronic band gap is `'indirect'`, the reciprocal momentum transfer for which the band gap is defined
        in units of the `reciprocal_lattice_vectors`. The initial and final momentum 3D vectors are given in the first
        and second element. Example, the momentum transfer in bulk Si2 happens between the Γ and the (approximately)
        X points in the Brillouin zone; thus:
            `momentum_transfer = [[0, 0, 0], [0.5, 0.5, 0]]`.

        Note: this quantity only refers to scalar `value`, not to arrays of `value`.
        """,
    )

    spin_channel = Quantity(
        type=np.int32,
        description="""
        Spin channel of the corresponding electronic band gap. It can take values of 0 or 1.
        """,
    )

    value = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        The value of the electronic band gap. This value has to be positive, otherwise it will
        prop an error and be set to None by the `normalize()` function.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name
        self.rank = []

    def validate_values(self, logger: 'BoundLogger') -> Optional[pint.Quantity]:
        """
        Validate the electronic band gap `value` by checking if they are negative and sets them to None if they are.

        Args:
            logger (BoundLogger): The logger to log messages.
        """
        value = self.value.magnitude
        if not isinstance(self.value.magnitude, np.ndarray):  # for scalars
            value = np.array(
                [value]
            )  # ! check this when talking with Lauri and Theodore

        # Set the value to 0 when it is negative
        if (value < 0).any():
            logger.error('The electronic band gap cannot be defined negative.')
            return None

        if not isinstance(self.value.magnitude, np.ndarray):  # for scalars
            value = value[0]
        return value * self.value.u

    def resolve_type(self, logger: 'BoundLogger') -> Optional[str]:
        """
        Resolves the `type` of the electronic band gap based on the stored `momentum_transfer` values.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[str]): The resolved `type` of the electronic band gap.
        """
        mtr = self.momentum_transfer if self.momentum_transfer is not None else []

        # Check if the `momentum_transfer` is [], and return the type and a warning in the log for `indirect` band gaps
        if len(mtr) == 0:
            if self.type == 'indirect':
                logger.warning(
                    'The `momentum_transfer` is not stored for an `indirect` band gap.'
                )
            return self.type

        # Check if the `momentum_transfer` has at least two elements, and return None if it does not
        if len(mtr) == 1:
            logger.warning(
                'The `momentum_transfer` should have at least two elements so that the difference can be calculated and the type of electronic band gap can be resolved.'
            )
            return None

        # Resolve `type` from the difference between the initial and final momentum transfer
        momentum_difference = np.diff(mtr, axis=0)
        if (np.isclose(momentum_difference, np.zeros(3))).all():
            return 'direct'
        else:
            return 'indirect'

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Checks if the `value` is negative and sets it to None if it is.
        self.value = self.validate_values(logger)
        if self.value is None:
            # ? What about deleting the class if `value` is None?
            logger.error('The `value` of the electronic band gap is not stored.')
            return

        # Resolve the `type` of the electronic band gap from `momentum_transfer`, ONLY for scalar `value`
        if isinstance(self.value.magnitude, np.ndarray):
            logger.info(
                'We do not support `type` which describe individual elements in an array `value`.'
            )
        else:
            self.type = self.resolve_type(logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING

import numpy as np

from nomad.metainfo import Quantity

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import PhysicalProperty


class FermiLevel(PhysicalProperty):
    """
    Energy required to add or extract a charge from a material at zero temperature. It can be also defined as the chemical potential at zero temperature.
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    iri = 'http://fairmat-nfdi.eu/taxonomy/FermiLevel'

    value = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        The value of the Fermi level.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = []
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class ChemicalPotential(PhysicalProperty):
    """
    Free energy cost of adding or extracting a particle from a thermodynamic system.
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    iri = 'http://fairmat-nfdi.eu/taxonomy/ChemicalPotential'

    value = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        The value of the chemical potential.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = []
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np
import pint

from nomad.config import config
from nomad.metainfo import MEnum, Quantity, SubSection

if TYPE_CHECKING:
    from nomad.metainfo import Section, Context
    from nomad.datamodel.datamodel import EntryArchive
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.atoms_state import AtomsState, OrbitalsState
from nomad_simulations.schema_packages.physical_property import PhysicalProperty
from nomad_simulations.schema_packages.properties.band_gap import ElectronicBandGap
from nomad_simulations.schema_packages.utils import get_sibling_section, get_variables
from nomad_simulations.schema_packages.variables import Energy2 as Energy

configuration = config.get_plugin_entry_point(
    'nomad_simulations.schema_packages:nomad_simulations_plugin'
)


class SpectralProfile(PhysicalProperty):
    """
    A base section used to define the spectral profile.
    """

    value = Quantity(
        type=np.float64,
        description="""
        The value of the intensities of a spectral profile in arbitrary units.
        """,
    )  # TODO check units and normalization_factor of DOS and Spectras and see whether they can be merged

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = []

    def is_valid_spectral_profile(self) -> bool:
        """
        Check if the spectral profile is valid, i.e., if all `value` are defined positive.

        Returns:
            (bool): True if the spectral profile is valid, False otherwise.
        """
        if (self.value < 0.0).any():
            return False
        return True

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        if self.is_valid_spectral_profile() is False:
            logger.error(
                'Invalid negative intensities found: could not validate spectral profile.'
            )
            return


class DOSProfile(SpectralProfile):
    """
    A base section used to define the `value` of the `ElectronicDensityOfState` property. This is useful when containing
    contributions for `projected_dos` with the correct unit.
    """

    value = Quantity(
        type=np.float64,
        unit='1/joule',
        description="""
        The value of the electronic DOS.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)

    def resolve_pdos_name(self, logger: 'BoundLogger') -> Optional[str]:
        """
        Resolve the `name` of the projected `DOSProfile` from the `entity_ref` section. This is resolved as:
            - `'atom X'` with 'X' being the chemical symbol for `AtomsState` references.
            -  `'orbital Y X'` with 'X' being the chemical symbol and 'Y' the orbital label for `OrbitalsState` references.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[str]): The resolved `name` of the projected DOS profile.
        """
        if self.entity_ref is None:
            logger.warning(
                'The `entity_ref` is not set for the DOS profile. Could not resolve the `name`.'
            )
            return None

        # Resolve the `name` from the `entity_ref`
        name = None
        if isinstance(self.entity_ref, AtomsState):
            name = f'atom {self.entity_ref.chemical_symbol}'
        elif isinstance(self.entity_ref, OrbitalsState):
            name = f'orbital {self.entity_ref.l_quantum_symbol}{self.entity_ref.ml_quantum_symbol} {self.entity_ref.m_parent.chemical_symbol}'
        return name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # We resolve
        self.name = self.resolve_pdos_name(logger)


class ElectronicDensityOfStates(DOSProfile):
    """
    Number of electronic states accessible for the charges per energy and per volume.
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`
    iri = 'http://fairmat-nfdi.eu/taxonomy/ElectronicDensityOfStates'

    spin_channel = Quantity(
        type=np.int32,
        description="""
        Spin channel of the corresponding electronic DOS. It can take values of 0 or 1.
        """,
    )

    # TODO clarify the role of `energies_origin` once `ElectronicEigenvalues` is implemented
    energies_origin = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        Energy level denoting the origin along the energy axis, used for comparison and visualization. It is
        defined as the `ElectronicEigenvalues.highest_occupied_energy`.
        """,
    )

    normalization_factor = Quantity(
        type=np.float64,
        description="""
        Normalization factor for electronic DOS to get a cell-independent intensive DOS. The cell-independent
        intensive DOS is as the integral from the lowest (most negative) energy to the Fermi level for a neutrally
        charged system (i.e., the sum of `AtomsState.charge` is zero).
        """,
    )

    # ? Do we want to store the integrated value here os as part of an nomad-analysis tool? Check `dos_integrator.py` module in dos normalizer repository
    # value_integrated = Quantity(
    #     type=np.float64,
    #     description="""
    #     The cumulative intensities integrated from from the lowest (most negative) energy to the Fermi level.
    #     """,
    # )

    projected_dos = SubSection(
        sub_section=DOSProfile.m_def,
        repeats=True,
        description="""
        Projected DOS. It can be atom- (different elements in the unit cell) or orbital-projected. These can be calculated in a cascade as:
            - If the total DOS is not present, we sum all atom-projected DOS to obtain it.
            - If the atom-projected DOS is not present, we sum all orbital-projected DOS to obtain it.
        Note: the cover given by summing up contributions is not perfect, and will depend on the projection functions used.

        In `projected_dos`, `name` and `entity_ref` must be set in order for normalization to work:
            - The `entity_ref` is the `OrbitalsState` or `AtomsState` sections.
            - The `name` of the projected DOS should be `'atom X'` or `'orbital Y X'`, with 'X' being the chemical symbol and 'Y' the orbital label.
            These can be extracted from `entity_ref`.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def resolve_energies_origin(
        self,
        energies: pint.Quantity,
        fermi_level: Optional[pint.Quantity],
        logger: 'BoundLogger',
    ) -> Optional[pint.Quantity]:
        """
        Resolve the origin of reference for the energies from the sibling `ElectronicEigenvalues` section and its
        `highest_occupied` level, or if this does not exist, from the `fermi_level` value as extracted from the sibling property, `FermiLevel`.

        Args:
            fermi_level (Optional[pint.Quantity]): The resolved Fermi level.
            energies (pint.Quantity): The grid points of the `Energy` variable.
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[pint.Quantity]): The resolved origin of reference for the energies.
        """
        # Check if the variables contain more than one variable (different than Energy)
        # ? Is this correct or should be use the index of energies to extract the proper shape element in `self.value` being used for `dos_values`?
        if len(self.variables) > 1:
            logger.warning(
                'The ElectronicDensityOfStates section contains more than one variable. We cannot extract the energy reference.'
            )
            return None

        # Extract the `ElectronicEigenvalues` section to get the `highest_occupied` and `lowest_unoccupied` energies
        # TODO implement once `ElectronicEigenvalues` is in the schema
        eigenvalues = get_sibling_section(
            section=self, sibling_section_name='electronic_eigenvalues', logger=logger
        )  # we consider `index_sibling` to be 0
        highest_occupied_energy = (
            eigenvalues.highest_occupied if eigenvalues is not None else None
        )
        lowest_unoccupied_energy = (
            eigenvalues.lowest_unoccupied if eigenvalues is not None else None
        )
        # and set defaults for `highest_occupied_energy` and `lowest_unoccupied_energy` in `m_cache`
        if highest_occupied_energy is not None:
            self.m_cache['highest_occupied_energy'] = highest_occupied_energy
        if lowest_unoccupied_energy is not None:
            self.m_cache['lowest_unoccupied_energy'] = lowest_unoccupied_energy

        # Check that the closest `energies` to the energy reference is not too far away.
        # If it is very far away, normalization may be very inaccurate and we do not report it.
        dos_values = self.value.magnitude
        eref = highest_occupied_energy if fermi_level is None else fermi_level
        fermi_idx = (np.abs(energies - eref)).argmin()
        fermi_energy_closest = energies[fermi_idx]
        distance = np.abs(fermi_energy_closest - eref)
        single_peak_fermi = False
        if distance.magnitude <= configuration.dos_energy_tolerance:
            # See if there are zero values close below the energy reference.
            idx = fermi_idx
            idx_descend = fermi_idx
            while True:
                try:
                    value = dos_values[idx]
                    energy_distance = np.abs(eref - energies[idx])
                except IndexError:
                    break
                if energy_distance.magnitude > configuration.dos_energy_tolerance:
                    break
                if value <= configuration.dos_intensities_threshold:
                    idx_descend = idx
                    break
                idx -= 1

            # See if there are zero values close above the fermi energy.
            idx = fermi_idx
            idx_ascend = fermi_idx
            while True:
                try:
                    value = dos_values[idx]
                    energy_distance = np.abs(eref - energies[idx])
                except IndexError:
                    break
                if energy_distance.magnitude > configuration.dos_energy_tolerance:
                    break
                if value <= configuration.dos_intensities_threshold:
                    idx_ascend = idx
                    break
                idx += 1

            # If there is a single peak at fermi energy, no
            # search needs to be performed.
            if idx_ascend != fermi_idx and idx_descend != fermi_idx:
                self.m_cache['highest_occupied_energy'] = fermi_energy_closest
                self.m_cache['lowest_unoccupied_energy'] = fermi_energy_closest
                single_peak_fermi = True

            if not single_peak_fermi:
                # Look for highest occupied energy below the descend index
                idx = idx_descend
                while True:
                    try:
                        value = dos_values[idx]
                    except IndexError:
                        break
                    if value > configuration.dos_intensities_threshold:
                        idx = idx if idx == idx_descend else idx + 1
                        self.m_cache['highest_occupied_energy'] = energies[idx]
                        break
                    idx -= 1
                # Look for lowest unoccupied energy above idx_ascend
                idx = idx_ascend
                while True:
                    try:
                        value = dos_values[idx]
                    except IndexError:
                        break
                    if value > configuration.dos_intensities_threshold:
                        idx = idx if idx == idx_ascend else idx - 1
                        self.m_cache['highest_occupied_energy'] = energies[idx]
                        break
                    idx += 1

        # Return the `highest_occupied_energy` as the `energies_origin`, or the `fermi_level` if it is not None
        energies_origin = self.m_cache.get('highest_occupied_energy')
        if energies_origin is None:
            energies_origin = fermi_level
        return energies_origin

    def resolve_normalization_factor(self, logger: 'BoundLogger') -> Optional[float]:
        """
        Resolve the `normalization_factor` for the electronic DOS to get a cell-independent intensive DOS.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[float]): The normalization factor.
        """
        # Get the `ModelSystem` as referenced in the `Outputs.model_system_ref`
        model_system = get_sibling_section(
            section=self, sibling_section_name='model_system_ref', logger=logger
        )
        if model_system is None:
            logger.warning(
                'Could not resolve the referenced `ModelSystem` in the `Outputs`.'
            )
            return None

        # Get the originally parsed `AtomicCell`, which is the first element stored in `ModelSystem.cell` of name `'AtomicCell'`
        atomic_cell = None
        for cell in model_system.cell:
            if cell.name == 'AtomicCell':  # we get the originally parsed `AtomicCell`
                atomic_cell = cell
                break
        if atomic_cell is None:
            logger.warning(
                'Could not resolve the `AtomicCell` from the referenced `ModelSystem`.'
            )
            return None

        # Get the `atoms_state` and their `atomic_number` from the `AtomicCell`
        if atomic_cell.atoms_state is None or len(atomic_cell.atoms_state) == 0:
            logger.warning('Could not resolve the `atoms_state` from the `AtomicCell`.')
            return None
        atomic_numbers = [atom.atomic_number for atom in atomic_cell.atoms_state]

        # Return `normalization_factor` depending if the calculation is spin polarized or not
        if self.spin_channel is not None:
            normalization_factor = 1 / (2 * sum(atomic_numbers))
        else:
            normalization_factor = 1 / sum(atomic_numbers)
        return normalization_factor

    def extract_band_gap(self) -> Optional[ElectronicBandGap]:
        """
        Extract the electronic band gap from the `highest_occupied_energy` and `lowest_unoccupied_energy` stored
        in `m_cache` from `resolve_energies_origin()`. If the difference of `highest_occupied_energy` and
        `lowest_unoccupied_energy` is negative, the band gap `value` is set to 0.0.

        Returns:
            (Optional[ElectronicBandGap]): The extracted electronic band gap section to be stored in `Outputs`.
        """
        band_gap = None
        homo = self.m_cache.get('highest_occupied_energy')
        lumo = self.m_cache.get('lowest_unoccupied_energy')
        if homo and lumo:
            band_gap = ElectronicBandGap()
            band_gap.is_derived = True
            band_gap.physical_property_ref = self

            if (homo - lumo).magnitude < 0:
                band_gap.value = 0.0
            else:
                band_gap.value = homo - lumo
        return band_gap

    def extract_projected_dos(
        self, type: str, logger: 'BoundLogger'
    ) -> List[Optional[DOSProfile]]:
        """
        Extract the projected DOS from the `projected_dos` section and the specified `type`.

        Args:
            type (str): The type of the projected DOS to extract. It can be `'atom'` or `'orbital'`.

        Returns:
            (DOSProfile): The extracted projected DOS.
        """
        extracted_pdos = []
        for pdos in self.projected_dos:
            # We make sure each PDOS is normalized
            pdos.normalize(None, logger)

            # Initial check for `name` and `entity_ref`
            if (
                pdos.name is None
                or pdos.entity_ref is None
                or len(pdos.entity_ref) == 0
            ):
                logger.warning(
                    '`name` or `entity_ref` are not set for `projected_dos` and they are required for normalization to work.'
                )
                return None

            if type in pdos.name:
                extracted_pdos.append(pdos)
        return extracted_pdos

    def generate_from_projected_dos(
        self, logger: 'BoundLogger'
    ) -> Optional[pint.Quantity]:
        """
        Generate the total `value` of the electronic DOS from the `projected_dos` contributions. If the `projected_dos`
        is not present, it returns `None`.

        Args:
            logger (BoundLogger): The logger to log messages.

        Returns:
            (Optional[pint.Quantity]): The total `value` of the electronic DOS.
        """
        if self.projected_dos is None or len(self.projected_dos) == 0:
            return None

        # Extract `Energy` variables
        energies = get_variables(self.variables, Energy)
        if len(energies) != 1:
            logger.warning(
                'The `ElectronicDensityOfStates` does not contain an `Energy` variable to extract the DOS.'
            )
            return None

        # We distinguish between orbital and atom `projected_dos`
        orbital_projected = self.extract_projected_dos('orbital', logger)
        atom_projected = self.extract_projected_dos('atom', logger)

        # Extract `atom_projected` from `orbital_projected` by summing up the `orbital_projected` contributions for each atom
        if len(atom_projected) == 0:
            atom_data: Dict[AtomsState, List[DOSProfile]] = {}
            for orb_pdos in orbital_projected:
                # `entity_ref` is the `OrbitalsState` section, whose parent is `AtomsState`
                entity_ref = orb_pdos.entity_ref.m_parent
                if entity_ref in atom_data:
                    atom_data[entity_ref].append(orb_pdos)
                else:
                    atom_data[entity_ref] = [orb_pdos]
            for ref, data in atom_data.items():
                atom_dos = DOSProfile(
                    name=f'atom {ref.chemical_symbol}',
                    entity_ref=ref,
                    variables=energies,
                )
                orbital_values = [
                    dos.value.magnitude for dos in data
                ]  # to avoid warnings from pint
                orbital_unit = data[0].value.u
                atom_dos.value = np.sum(orbital_values, axis=0) * orbital_unit
                atom_projected.append(atom_dos)
            # We concatenate the `atom_projected` to the `projected_dos`
            self.projected_dos = orbital_projected + atom_projected

        # Extract `value` from `atom_projected` by summing up the `atom_projected` contributions
        value = self.value
        if value is None:
            atom_values = [
                dos.value.magnitude for dos in atom_projected
            ]  # to avoid warnings from pint
            atom_unit = atom_projected[0].value.u
            value = np.sum(atom_values, axis=0) * atom_unit
        return value

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        # Initial check to see if `variables` contains the required `Energy` variable
        energies = get_variables(self.variables, Energy)
        if len(energies) != 0:
            return
        energies = energies[0].points

        # Resolve `fermi_level` from a sibling section with respect to `ElectronicDensityOfStates`
        fermi_level = get_sibling_section(
            section=self, sibling_section_name='fermi_level', logger=logger
        )  # * we consider `index_sibling` to be 0
        if fermi_level is not None:
            fermi_level = fermi_level.value
        # and the `energies_origin` from the sibling `ElectronicEigenvalues` section
        self.energies_origin = self.resolve_energies_origin(
            energies, fermi_level, logger
        )
        if self.energies_origin is None:
            logger.info('Could not resolve the `energies_origin` for the DOS')

        # Resolve `normalization_factor`
        if self.normalization_factor is None:
            self.normalization_factor = self.resolve_normalization_factor(logger)

        # `ElectronicBandGap` extraction
        band_gap = self.extract_band_gap()
        if band_gap is not None:
            self.m_parent.electronic_band_gap.append(band_gap)

        # Total `value` extraction from `projected_dos`
        value_from_pdos = self.generate_from_projected_dos(logger)
        if self.value is None and value_from_pdos is not None:
            logger.info(
                'The `ElectronicDensityOfStates.value` is not stored. We will attempt to obtain it by summing up projected DOS contributions, if these are present.'
            )
            self.value = value_from_pdos


class AbsorptionSpectrum(SpectralProfile):
    """ """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    axis = Quantity(
        type=MEnum('xx', 'yy', 'zz'),
        description="""
        Axis of the absorption spectrum. This is related with the polarization direction, and can be seen as the
        principal term in the tensor `Permittivity.value` (see permittivity.py module).
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        # Set the name of the section
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class XASSpectrum(AbsorptionSpectrum):
    """
    X-ray Absorption Spectrum (XAS).
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    xanes_spectrum = SubSection(
        sub_section=AbsorptionSpectrum.m_def,
        description="""
        X-ray Absorption Near Edge Structure (XANES) spectrum.
        """,
        repeats=False,
    )

    exafs_spectrum = SubSection(
        sub_section=AbsorptionSpectrum.m_def,
        description="""
        Extended X-ray Absorption Fine Structure (EXAFS) spectrum.
        """,
        repeats=False,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        # Set the name of the section
        self.name = self.m_def.name

    def generate_from_contributions(self, logger: 'BoundLogger') -> None:
        """
        Generate the `value` of the XAS spectrum by concatenating the XANES and EXAFS contributions. It also concatenates
        the `Energy` grid points of the XANES and EXAFS parts.

        Args:
            logger (BoundLogger): The logger to log messages.
        """
        # TODO check if this method is general enough
        if self.xanes_spectrum is not None and self.exafs_spectrum is not None:
            # Concatenate XANE and EXAFS `Energy` grid points
            xanes_variables = get_variables(self.xanes_spectrum.variables, Energy)
            exafs_variables = get_variables(self.exafs_spectrum.variables, Energy)
            if len(xanes_variables) == 0 or len(exafs_variables) == 0:
                logger.warning(
                    'Could not extract the `Energy` grid points from XANES or EXAFS.'
                )
                return
            xanes_energies = xanes_variables[0].points
            exafs_energies = exafs_variables[0].points
            if xanes_energies.max() > exafs_energies.min():
                logger.warning(
                    'The XANES `Energy` grid points are not below the EXAFS `Energy` grid points.'
                )
                return
            self.variables = [
                Energy(points=np.concatenate([xanes_energies, exafs_energies]))
            ]
            # Concatenate XANES and EXAFS `value` if they have the same shape ['n_energies']
            try:
                self.value = np.concatenate(
                    [self.xanes_spectrum.value, self.exafs_spectrum.value]
                )
            except ValueError:
                logger.warning(
                    'The XANES and EXAFS `value` have different shapes. Could not concatenate the values.'
                )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        if self.value is None:
            logger.info(
                'The `XASSpectrum.value` is not stored. We will attempt to obtain it by combining the XANES and EXAFS parts if these are present.'
            )
            self.generate_from_contributions(logger)
